from __future__ import annotations

import copy
import json
import logging
import queue
from typing import Any, Dict, List, Optional, Text

from apscheduler.schedulers.background import BackgroundScheduler

from rasa.anonymization.anonymisation_rule_yaml_reader import (
    AnonymizationRulesYamlReader,
)
from rasa.anonymization.anonymization_rule_executor import AnonymizationRuleList
from rasa.anonymization.anonymization_rule_orchestrator import (
    AnonymizationRuleOrchestrator,
)
from rasa.core.brokers.kafka import KafkaEventBroker
from rasa.shared.core.events import Event
from rasa.utils.endpoints import EndpointConfig, read_endpoint_config
from rasa.utils.singleton import Singleton

logger = logging.getLogger(__name__)


class AnonymizationPipeline:
    def run(self, event: Dict[Text, Any]) -> None:
        """Run the anonymization pipeline on the given event.

        Args:
            event: The event to anonymize
        """
        ...

    def log_run(self, data: Any) -> Any:
        """Anonymize the log data.

        Args:
            data: log data to anonymize

        Returns:
            Anonymized log data
        """
        ...

    def stop(self) -> None:
        """Stop the anonymization pipeline."""
        ...


class BackgroundAnonymizationPipeline(AnonymizationPipeline):
    EVENT_QUEUE_PROCESSING_TIMEOUT_IN_SECONDS = 2.0

    def __init__(self, anonymization_pipeline: AnonymizationPipeline):
        self.anonymization_pipeline = anonymization_pipeline

        # Order of the initialisation is important
        # The event queue must be created before the scheduler
        # The can_consume_event_queue must be set to True before the scheduler starts
        self.event_queue: queue.Queue = queue.Queue()

        # This flag is used to stop the scheduler
        self.can_consume_from_event_queue = True
        self.event_anonymisation_scheduler = BackgroundScheduler()
        self.event_anonymisation_scheduler.add_job(
            self._consumer_queue, max_instances=1
        )
        self.event_anonymisation_scheduler.start()

    def stop(self) -> None:
        logger.debug("Shutting down the anonymization pipeline...")
        self.can_consume_from_event_queue = False
        self.event_anonymisation_scheduler.shutdown()

    def run(self, event: Dict[Text, Any]) -> None:
        self.event_queue.put(event)

    def log_run(self, data: Any) -> Any:
        return self.anonymization_pipeline.log_run(data)

    def _consumer_queue(self) -> None:
        while self.can_consume_from_event_queue:
            try:
                # Wait for 2 seconds for an event to be added to the queue
                # If no event is added to the queue, continue
                # This is done to avoid the scheduler to be stuck in the while loop
                # when we want to stop the scheduler
                event = self.event_queue.get(
                    timeout=self.EVENT_QUEUE_PROCESSING_TIMEOUT_IN_SECONDS
                )
                self.anonymization_pipeline.run(event)
                self.event_queue.task_done()
            except queue.Empty:
                continue


class SyncAnonymizationPipeline(AnonymizationPipeline):
    """Pipeline for anonymizing events."""

    def __init__(self, orchestrators: List[AnonymizationRuleOrchestrator]) -> None:
        """Initializes the pipeline."""
        self.orchestrators = orchestrators

    def stop(self) -> None:
        pass

    def run(self, event: Dict[Text, Any]) -> None:
        """Runs the anonymization pipeline."""
        logger.debug("Running the anonymization pipeline for event...")

        for orchestrator in self.orchestrators:
            anonymized_event = orchestrator.anonymize_event(event)
            is_anonymized = True if event != anonymized_event else False
            orchestrator.publish_event(anonymized_event, is_anonymized)

    def log_run(self, data: Any) -> Any:
        """Runs the anonymization pipeline for logging."""
        logger.debug("Running the anonymization pipeline for logs....")

        anonymized_data = None

        # this is to make sure that the original data is not modified
        data_copy = copy.deepcopy(data)

        for orchestrator in self.orchestrators:
            # orchestrator for anonymizing logs has its event broker set to None
            if isinstance(orchestrator.event_broker, KafkaEventBroker):
                continue

            if isinstance(data_copy, str):
                anonymized_data = orchestrator.anonymize_log_message(data_copy)

            elif isinstance(data_copy, list):
                anonymized_data = [
                    Event.from_parameters(orchestrator.anonymize_event(item.as_dict()))
                    if isinstance(item, Event)
                    else orchestrator.anonymize_log_message(str(item["value"]))
                    if isinstance(item, dict) and item.get("value") is not None
                    else item
                    for item in data_copy
                ]

            elif isinstance(data_copy, dict) and "event" in data_copy:
                anonymized_data = orchestrator.anonymize_event(data_copy)

            elif isinstance(data_copy, dict):
                anonymized_data = {}

                for key, value in data_copy.items():
                    try:
                        serialized_value = json.dumps(value)
                    except TypeError as error:
                        logger.error(
                            f"Failed to serialize value of type '{type(value)}' "
                            f"for key '{key}' before anonymization. "
                            f"Encountered error: {error}. "
                            f"Setting value to None."
                        )
                        serialized_value = None

                    anonymized_data[key] = (
                        orchestrator.anonymize_log_message(serialized_value)
                        if serialized_value is not None
                        else None
                    )

            else:
                logger.debug("Unsupported data type for logging anonymization.")

            if anonymized_data:
                return anonymized_data
        return data


class AnonymizationPipelineProvider(metaclass=Singleton):
    """Represents a provider for anonymization pipeline."""

    anonymization_pipeline: Optional[AnonymizationPipeline] = None

    def register_anonymization_pipeline(self, pipeline: AnonymizationPipeline) -> None:
        """Register an anonymization pipeline.

        Args:
            pipeline: The anonymization pipeline to register.
        """
        self.anonymization_pipeline = pipeline

    def get_anonymization_pipeline(self) -> Optional[AnonymizationPipeline]:
        """Get the anonymization pipeline.

        Returns:
            The anonymization pipeline.
        """
        return self.anonymization_pipeline


def load_anonymization_pipeline(endpoints_file: Optional[Text]) -> None:
    """Creates an anonymization pipeline."""
    yaml_reader = AnonymizationRulesYamlReader(endpoints_filename=endpoints_file)
    anonymization_rules = yaml_reader.read_anonymization_rules()

    if anonymization_rules is None:
        return None

    event_broker_config = read_endpoint_config(
        endpoints_file, endpoint_type="event_broker"
    )
    logging_config = read_endpoint_config(endpoints_file, endpoint_type="logger")

    orchestrators = _load_orchestrators(
        logging_config, event_broker_config, anonymization_rules
    )

    if not orchestrators:
        return None

    pipeline = SyncAnonymizationPipeline(orchestrators)
    async_anonymization_pipeline = BackgroundAnonymizationPipeline(pipeline)
    provider = AnonymizationPipelineProvider()
    provider.register_anonymization_pipeline(async_anonymization_pipeline)

    return None


def create_event_broker(
    topic_name: Text, event_broker_config: EndpointConfig
) -> Optional[KafkaEventBroker]:
    """Create a KafkaEventBroker object.

    Returns None if the event broker config is not of type 'kafka'.
    """
    if event_broker_config.type != "kafka":
        logger.warning(
            f"Unsupported event broker config provided. "
            f"Expected type 'kafka' but got "
            f"'{event_broker_config.type}'. "
            f"Setting event broker to None."
        )
        event_broker = None
    else:
        logger.debug(f"Setting topic to '{topic_name}'.")
        event_broker_config.kwargs["topic"] = topic_name
        event_broker = KafkaEventBroker(
            event_broker_config.url, **event_broker_config.kwargs
        )

    return event_broker


def _load_orchestrators(
    logging_config: EndpointConfig,
    event_broker_config: EndpointConfig,
    anonymization_rules: List[AnonymizationRuleList],
) -> List[AnonymizationRuleOrchestrator]:
    orchestrators = []

    if logging_config:
        formatter = logging_config.kwargs.get("formatter", "")
        logging_rule = formatter.get("anonymization_rules")

        for rule_list in anonymization_rules:
            if rule_list.id == logging_rule:
                orchestrators.append(AnonymizationRuleOrchestrator(None, rule_list))

    if event_broker_config:
        anonymization_topics = event_broker_config.kwargs.get(
            "anonymization_topics", []
        )
        visited_topics = []

        for rule_list in anonymization_rules:
            for topic in anonymization_topics:
                topic_name = topic.get("name")

                if topic_name in visited_topics:
                    continue

                if rule_list.id == topic.get("anonymization_rules"):
                    event_broker = create_event_broker(topic_name, event_broker_config)

                    orchestrators.append(
                        AnonymizationRuleOrchestrator(event_broker, rule_list)
                    )

                    visited_topics.append(topic_name)
    return orchestrators



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional, Text

from rasa.dialogue_understanding.stack.frames import PatternFlowStackFrame
from rasa.shared.constants import (
    RASA_DEFAULT_FLOW_PATTERN_PREFIX,
    RASA_PATTERN_CANNOT_HANDLE_DEFAULT,
)

FLOW_PATTERN_CANNOT_HANDLE = RASA_DEFAULT_FLOW_PATTERN_PREFIX + "cannot_handle"


@dataclass
class CannotHandlePatternFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame that gets added when that the
    bot can't handle the user's input."""

    flow_id: str = FLOW_PATTERN_CANNOT_HANDLE
    """The ID of the flow."""

    reason: Optional[Text] = RASA_PATTERN_CANNOT_HANDLE_DEFAULT
    """Reason for cannot handle used in switch-case of the
    cannot handle pattern flow."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_CANNOT_HANDLE

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> CannotHandlePatternFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return CannotHandlePatternFlowStackFrame(
            frame_id=data["frame_id"], step_id=data["step_id"], reason=data["reason"]
        )



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import structlog

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    FlowStackFrameType,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.utils import (
    top_user_flow_frame,
    user_flows_on_the_stack,
)
from rasa.shared.core.events import Event, FlowInterrupted
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


@dataclass
class StartFlowCommand(Command):
    """A command to start a flow."""

    flow: str

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "start flow"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> StartFlowCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        try:
            return StartFlowCommand(flow=data["flow"])
        except KeyError as e:
            raise ValueError(
                f"Missing parameter '{e}' while parsing StartFlowCommand."
            ) from e

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        original_stack = original_tracker.stack
        applied_events: List[Event] = []

        if self.flow in user_flows_on_the_stack(stack):
            structlogger.debug(
                "command_executor.skip_command.already_started_flow", command=self
            )
            return []
        elif self.flow not in all_flows.flow_ids:
            structlogger.debug(
                "command_executor.skip_command.start_invalid_flow_id", command=self
            )
            return []

        original_user_frame = top_user_flow_frame(original_stack)
        original_top_flow = (
            original_user_frame.flow(all_flows) if original_user_frame else None
        )

        frame_type = FlowStackFrameType.REGULAR

        if original_top_flow:
            frame_type = FlowStackFrameType.INTERRUPT

            if original_user_frame is not None:
                applied_events.append(
                    FlowInterrupted(
                        original_user_frame.flow_id, original_user_frame.step_id
                    )
                )

        structlogger.debug("command_executor.start_flow", command=self)
        stack.push(UserFlowStackFrame(flow_id=self.flow, frame_type=frame_type))
        return applied_events + tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.flow)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StartFlowCommand):
            return False

        return other.flow == self.flow

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: f"StartFlow({self.flow})",
            CommandSyntaxVersion.v2: f"start flow {self.flow}",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> Optional[StartFlowCommand]:
        """Converts the DSL string to a command."""
        return StartFlowCommand(flow=str(match.group(1).strip()))

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: r"StartFlow\(['\"]?([a-zA-Z0-9_-]+)['\"]?\)",
            CommandSyntaxVersion.v2: (
                r"""^[\s\W\d]*start flow ['"`]?([a-zA-Z0-9_-]+)['"`]*"""
            ),
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )



from collections import defaultdict
from typing import Dict, List

from pydantic import BaseModel

from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding_test.command_comparison import (
    is_command_present_in_list,
)
from rasa.dialogue_understanding_test.du_test_result import (
    DialogueUnderstandingTestResult,
)


class CommandMetrics(BaseModel):
    tp: int
    fp: int
    fn: int
    total_count: int

    @staticmethod
    def _safe_divide(numerator: float, denominator: float) -> float:
        """Safely perform division, returning 0.0 if the denominator is zero."""
        return numerator / denominator if denominator > 0 else 0.0

    def get_precision(self) -> float:
        return self._safe_divide(self.tp, self.tp + self.fp)

    def get_recall(self) -> float:
        return self._safe_divide(self.tp, self.tp + self.fn)

    def get_f1_score(self) -> float:
        precision = self.get_precision()
        recall = self.get_recall()

        return self._safe_divide(2 * precision * recall, precision + recall)

    def as_dict(self) -> Dict[str, float]:
        return {
            "tp": self.tp,
            "fp": self.fp,
            "fn": self.fn,
            "precision": self.get_precision(),
            "recall": self.get_recall(),
            "f1_score": self.get_f1_score(),
            "total_count": self.total_count,
        }


def calculate_command_metrics(
    test_results: List[DialogueUnderstandingTestResult],
) -> Dict[str, CommandMetrics]:
    """Calculate the command metrics for the test result."""
    metrics: Dict[str, CommandMetrics] = defaultdict(
        lambda: CommandMetrics(tp=0, fp=0, fn=0, total_count=0)
    )

    for test_result in test_results:
        _increase_total_count(test_result.get_expected_commands(), metrics)

        # if the test case passed, count all commands as tp
        if test_result.passed:
            _increase_tp(test_result.get_expected_commands(), metrics)
            continue

        # in case the test case failed, we need to compare
        # the expected and actual commands for each step
        for step in test_result.test_case.iterate_over_user_steps():
            expected_commands = step.commands
            predicted_commands = step.get_predicted_commands()

            _update_metrics_true_positive_and_false_negative(
                expected_commands, predicted_commands, metrics
            )
            _update_metrics_false_positive(
                expected_commands, predicted_commands, metrics
            )

    return metrics


def _get_command_name(command: Command) -> str:
    return command.command().replace(" ", "_")


def _increase_total_count(
    commands: List[Command],
    metrics: Dict[str, CommandMetrics],
) -> None:
    for command in commands:
        metrics[_get_command_name(command)].total_count += 1


def _increase_tp(
    commands: List[Command],
    metrics: Dict[str, CommandMetrics],
) -> None:
    for command in commands:
        metrics[_get_command_name(command)].tp += 1


def _update_metrics_true_positive_and_false_negative(
    expected_commands: List[Command],
    predicted_commands: List[Command],
    metrics: Dict[str, CommandMetrics],
) -> None:
    for expected_command in expected_commands:
        command_name = _get_command_name(expected_command)
        if is_command_present_in_list(expected_command, predicted_commands):
            metrics[command_name].tp += 1
        else:
            metrics[command_name].fn += 1


def _update_metrics_false_positive(
    expected_commands: List[Command],
    predicted_commands: List[Command],
    metrics: Dict[str, CommandMetrics],
) -> None:
    for predicted_command in predicted_commands:
        if not is_command_present_in_list(predicted_command, expected_commands):
            metrics[_get_command_name(predicted_command)].fp += 1



import math
import os
import pathlib
import shutil
import sys
from functools import lru_cache
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import pandas as pd
import rich
import structlog
from rich.table import Table

import rasa.shared.data
import rasa.shared.utils.cli
import rasa.utils.io
from rasa.e2e_test.constants import (
    KEY_FIXTURES,
    KEY_METADATA,
    KEY_STUB_CUSTOM_ACTIONS,
    KEY_TEST_CASE,
    KEY_TEST_CASES,
    STATUS_FAILED,
    STATUS_PASSED,
    STUB_CUSTOM_ACTION_NAME_SEPARATOR,
)
from rasa.e2e_test.e2e_test_case import Fixture, Metadata, TestCase, TestSuite
from rasa.e2e_test.stub_custom_action import (
    StubCustomAction,
    get_stub_custom_action_key,
)
from rasa.e2e_test.utils.validation import (
    read_e2e_test_schema,
    validate_path_to_test_cases,
    validate_test_case,
)
from rasa.shared.utils.yaml import (
    is_key_in_yaml,
    parse_raw_yaml,
    validate_yaml_data_using_schema_with_assertions,
)
from rasa.utils.beta import BetaNotEnabledException, ensure_beta_feature_is_enabled

if TYPE_CHECKING:
    from rasa.e2e_test.aggregate_test_stats_calculator import AccuracyCalculation
    from rasa.e2e_test.e2e_test_result import TestResult


RASA_PRO_BETA_STUB_CUSTOM_ACTION_ENV_VAR_NAME = "RASA_PRO_BETA_STUB_CUSTOM_ACTION"

structlogger = structlog.get_logger()


def color_difference(diff: List[str]) -> Generator[str, None, None]:
    """Colorize the difference between two strings.

    Example:
        >>> color_difference(["+ Hello", "- World"])
        ["<ansigreen>+ Hello</ansigreen>", "<ansired>- World</ansired>"]

    Args:
        diff: List of lines of the diff.

    Returns:
    Generator of colored lines.
    """
    for line in diff:
        if line.startswith("+"):
            yield "[green3]" + line + "[/green3]"
        elif line.startswith("-"):
            yield "[red3]" + line + "[/red3]"
        elif line.startswith("^"):
            yield "[blue3]" + line + "[/blue3]"
        elif line.startswith("?"):
            yield "[grey37]" + line + "[/grey37]"
        else:
            yield line


def print_failed_case(fail: "TestResult") -> None:
    """Print the details of a failed test case.

    Example:
        >>> print_failed_case(TestResult(TestCase([TestStep()]), True,
        ...                  ["- Hello", "+ World"]))
        ---------------------- test in test.md failed ----------------------
        Mismatch starting at test.md:1:
        <ansired>- Hello</ansired>
        <ansigreen>+ World</ansigreen>
    """
    fail_headline = (
        f"'{fail.test_case.name}' in {fail.test_case.file_with_line()} failed"
    )
    rasa.shared.utils.cli.print_error(
        f"{rasa.shared.utils.cli.pad(fail_headline, char='-')}\n"
    )
    rasa.shared.utils.cli.print_error(
        f"Mismatch starting at {fail.test_case.file}:{fail.error_line}: \n"
    )
    if fail.difference:
        rich.print(("\n".join(color_difference(fail.difference))))

    if fail.assertion_failure:
        rasa.shared.utils.cli.print_error(
            f"Assertion type '{fail.assertion_failure.assertion.type()}' failed "
            f"with this error message: {fail.assertion_failure.error_message}\n"
        )
        rasa.shared.utils.cli.print_error("Actual events transcript:\n")
        rasa.shared.utils.cli.print_error(
            "\n".join(fail.assertion_failure.actual_events_transcript)
        )


def print_test_summary(failed: List["TestResult"]) -> None:
    """Print the summary of the test run.

    Example:
        >>> print_test_summary([TestResult(TestCase([TestStep()]), True,
        ...                  ["- Hello", "+ World"])])
        =================== short test summary info ===================
        FAILED test.md::test
    """
    rasa.shared.utils.cli.print_info(
        rasa.shared.utils.cli.pad("short test summary info")
    )

    for f in failed:
        rasa.shared.utils.cli.print_error(
            f"FAILED {f.test_case.file}::{f.test_case.name}"
        )


def print_final_line(
    passed: List["TestResult"], failed: List["TestResult"], has_failed: bool
) -> None:
    """Print the final line of the test output.

    Args:
        passed: List of passed test cases.
        failed: List of failed test cases.
        has_failed: Boolean, true if the test run has failed.
    """
    final_line_color = "green3" if not has_failed else "red3"

    width = shutil.get_terminal_size((80, 20)).columns

    # calculate the length of the text - this is a bit hacky but works
    text_lengt = (
        math.ceil(len(passed) / 10)  # length of the number of passed tests
        + math.ceil(len(failed) / 10)  # length of the number of failed tests
        + 18  # length of the text "  failed, passed  "
    )
    # we can't use the padding function here as the text contains html tags
    # which are not taken into account when calculating the length
    padding = max(6, width - text_lengt)
    pre_pad = "=" * max(3, padding // 2)
    post_pad = "=" * max(3, math.ceil(padding / 2))
    rich.print(
        f"[{final_line_color}]{pre_pad} "
        f"[bold red3]{len(failed)} failed[/bold red3]"
        f"[bright_white], [/bright_white]"
        f"[bold green3]{len(passed)} passed[/bold green3]"
        f" {post_pad}[/{final_line_color}]"
    )


def print_aggregate_stats(accuracy_calculations: List["AccuracyCalculation"]) -> None:
    """Print the aggregate statistics of the test run."""
    rasa.shared.utils.cli.print_info(
        rasa.shared.utils.cli.pad("Accuracy By Assertion Type")
    )
    table = Table()
    table.add_column("Assertion Type", justify="center", style="cyan")
    table.add_column("Accuracy", justify="center", style="cyan")

    for accuracy_calculation in accuracy_calculations:
        table.add_row(
            accuracy_calculation.assertion_type, f"{accuracy_calculation.accuracy:.2%}"
        )

    rich.print(table)


def print_test_result(
    passed: List["TestResult"],
    failed: List["TestResult"],
    fail_fast: bool = False,
    **kwargs: Any,
) -> None:
    """Print the result of the test run.

    Args:
        passed: List of passed test cases.
        failed: List of failed test cases.
        fail_fast: If true, stop after the first failure.
        **kwargs: additional arguments
    """
    if failed:
        # print failure headline
        print("\n")
        rich.print(f"[bold]{rasa.shared.utils.cli.pad('FAILURES', char='=')}[/bold]")

    # print failed test_Case
    for fail in failed:
        print_failed_case(fail)

    accuracy_calculations = kwargs.get("accuracy_calculations", [])
    if accuracy_calculations:
        print_aggregate_stats(accuracy_calculations)

    print_test_summary(failed)

    if fail_fast:
        rasa.shared.utils.cli.print_error(
            rasa.shared.utils.cli.pad("stopping after 1 failure", char="!")
        )
        has_failed = True
    elif len(failed) + len(passed) == 0:
        # no tests were run, print error
        rasa.shared.utils.cli.print_error(
            rasa.shared.utils.cli.pad("no test cases found", char="!")
        )
        print_e2e_help()
        has_failed = True
    elif failed:
        has_failed = True
    else:
        has_failed = False

    print_final_line(passed, failed, has_failed=has_failed)
    sys.exit(1 if has_failed else 0)


def print_e2e_help() -> None:
    """Print help guiding users how to write e2e tests."""
    rasa.shared.utils.cli.print_info(
        dedent(
            """\
        To start using e2e tests create a yaml file in a test directory, e.g.
        'tests/test_cases.yml'. You can find example test cases in the starter
        pack at

            https://github.com/RasaHQ/starter-pack-intentless-policy#testing-the-policy

        Here is an example of a test case in yaml format:

            test_cases:
            - test_case: "test_greet"
              steps:
              - user: "hello there!"
              - bot: "Hey! How are you?"

        To run the e2e tests, execute:
            >>> rasa test e2e <path-to-test-cases>
    """
        )
    )


def split_into_passed_failed(
    results: List["TestResult"],
) -> Tuple[List["TestResult"], List["TestResult"]]:
    """Get the summary of the test results.

    Args:
        results: List of test results.

    Returns:
        Tuple consisting of passed count, failed count and failed test cases.
    """
    passed_cases = [r for r in results if r.pass_status]
    failed_cases = [r for r in results if not r.pass_status]

    return passed_cases, failed_cases


@lru_cache(maxsize=1)
def extract_test_case_from_path(path: str) -> Tuple[str, str]:
    """Extract test case from path if specified.

    Args:
        path: Path to the file or folder containing test cases.

    Returns:
        Tuple consisting of the path to test cases and the extracted test case name.
    """
    test_case_name = ""

    if "::" in str(path):
        splitted_path = path.split("::")
        test_case_name = splitted_path[-1]
        path = splitted_path[0]

    return path, test_case_name


def is_test_case_file(file_path: Union[str, Path]) -> bool:
    """Check if file contains test cases.

    Args:
        file_path: Path of the file to check.

    Returns:
        `True` if the file contains test cases, `False` otherwise.
    """
    return rasa.shared.data.is_likely_yaml_file(file_path) and is_key_in_yaml(
        file_path, KEY_TEST_CASES
    )


def extract_test_cases(
    test_file_content: dict,
    test_case_name: str,
    test_file: str,
) -> List[TestCase]:
    """Extract test cases from the test file content.

    Args:
        test_file_content: Content of the test file.
        test_case_name: Name of the test case to extract.
        test_file: Path to the test file.

    Returns:
        List of test cases.
    """
    test_cases_content = test_file_content.get(KEY_TEST_CASES) or []
    if test_case_name:
        return [
            TestCase.from_dict(test_case_dict, file=test_file)
            for test_case_dict in test_cases_content
            if test_case_name == test_case_dict.get(KEY_TEST_CASE)
        ]
    return [
        TestCase.from_dict(test_case_dict, file=test_file)
        for test_case_dict in test_cases_content
    ]


def extract_fixtures(
    test_file_content: dict, existing_fixtures: Dict[str, Fixture]
) -> Dict[str, Fixture]:
    """Extract fixtures from the test file content.

    Args:
        test_file_content: Content of the test file.
        existing_fixtures: Existing fixtures.

    Returns:
        Dict of fixtures.
    """
    fixtures_content = test_file_content.get(KEY_FIXTURES) or []
    _fixtures = {}
    for fixture in fixtures_content:
        fixture_obj = Fixture.from_dict(fixture_dict=fixture)
        if existing_fixtures.get(fixture_obj.name) is None:
            _fixtures[fixture_obj.name] = fixture_obj
    return _fixtures


def extract_metadata(
    test_file_content: dict, existing_metadata: Dict[str, Metadata]
) -> Dict[str, Metadata]:
    """Extract metadata from the test file content.

    Args:
        test_file_content: Content of the test file.
        existing_metadata: Existing metadata.

    Returns:
        Dict of metadata.
    """
    metadata_contents = test_file_content.get(KEY_METADATA) or []
    _metadata = {}
    for metadata_content in metadata_contents:
        metadata_obj = Metadata.from_dict(metadata_dict=metadata_content)
        if existing_metadata.get(metadata_obj.name) is None:
            _metadata[metadata_obj.name] = metadata_obj
    return _metadata


def extract_stub_custom_actions(
    test_file_content: dict, test_file: str
) -> Dict[str, StubCustomAction]:
    """Extract stub custom actions from the test file content.

    Args:
        test_file_content: Content of the test file.
        test_file: Path to the test file.

    Returns:
        Dict of stub custom actions.
    """
    stub_custom_actions_contents = test_file_content.get(KEY_STUB_CUSTOM_ACTIONS) or {}
    _stub_custom_actions = {}
    for action_name, stub_data in stub_custom_actions_contents.items():
        if STUB_CUSTOM_ACTION_NAME_SEPARATOR in action_name:
            stub_custom_action_key = action_name
        else:
            test_file_name = Path(test_file).name
            stub_custom_action_key = get_stub_custom_action_key(
                test_file_name, action_name
            )
        _stub_custom_actions[stub_custom_action_key] = StubCustomAction.from_dict(
            action_name=action_name, stub_data=stub_data
        )
    return _stub_custom_actions


def read_test_cases(path: str) -> TestSuite:
    """Read test cases from the given path.

    Args:
        path: Path to the file or folder containing test cases.

    Returns:
        TestSuite.
    """
    # Extract test case path and name
    path, test_case_name = extract_test_case_from_path(path)
    validate_path_to_test_cases(path)

    # Load test files and schema
    test_files = rasa.shared.data.get_data_files([path], is_test_case_file)
    e2e_test_schema = read_e2e_test_schema()

    # Initialize containers
    input_test_cases = []
    fixtures: Dict[str, Fixture] = {}
    metadata: Dict[str, Metadata] = {}
    stub_custom_actions: Dict[str, StubCustomAction] = {}

    # Process each test file
    for test_file in test_files:
        test_file_content = parse_raw_yaml(Path(test_file).read_text(encoding="utf-8"))

        # Validate YAML content using the provided function
        validate_yaml_data_using_schema_with_assertions(
            yaml_data=test_file_content, schema_content=e2e_test_schema
        )

        # Parse test cases, fixtures, metadata, and stub custom actions
        test_cases = extract_test_cases(test_file_content, test_case_name, test_file)
        fixtures.update(extract_fixtures(test_file_content, fixtures))
        metadata.update(extract_metadata(test_file_content, metadata))
        stub_custom_actions.update(
            extract_stub_custom_actions(test_file_content, test_file)
        )
        input_test_cases.extend(test_cases)

    validate_test_case(test_case_name, input_test_cases, fixtures, metadata)
    if stub_custom_actions:
        check_beta_feature_flag_for_custom_actions_stubs()

    return TestSuite(
        input_test_cases,
        list(fixtures.values()),
        list(metadata.values()),
        stub_custom_actions,
    )


def check_beta_feature_flag_for_custom_actions_stubs() -> None:
    """Check the beta feature flag for custom actions stub and exit if not enabled."""
    try:
        ensure_beta_feature_is_enabled(
            "enabling stubs for custom actions",
            RASA_PRO_BETA_STUB_CUSTOM_ACTION_ENV_VAR_NAME,
        )
    except BetaNotEnabledException as exc:
        rasa.shared.utils.cli.print_error_and_exit(str(exc))


def _save_coverage_report(
    report: Optional[pd.DataFrame], test_status: str, output_dir: str
) -> None:
    if report is None:
        return

    if report is not None and not report.empty:
        if test_status == STATUS_PASSED:
            rasa.shared.utils.cli.print_success(report.to_string(index=False))
        else:
            rasa.shared.utils.cli.print_error(report.to_string(index=False))

    output_filename = f"coverage_report_for_{test_status}_tests.csv"
    output_file_path = os.path.join(output_dir, output_filename)
    report.to_csv(output_file_path, index=False)
    structlogger.info(
        "rasa.e2e_test.save_coverage_report",
        message=f"Coverage result for {test_status} e2e tests"
        f" is written to '{output_file_path}'.",
    )


def write_test_results_to_file(results: List["TestResult"], output_file: str) -> None:
    """Write test results to a file.

    Args:
        results: List of test results.
        output_file: Path to the output file.
    """
    Path(output_file).touch()

    data = {"test_results": [test_result.as_dict() for test_result in results]}

    rasa.utils.io.write_yaml(
        data, target=output_file, transform=transform_results_output_to_yaml
    )

    if STATUS_PASSED in output_file:
        rasa.shared.utils.cli.print_info(
            f"Passing test results have been saved at path: {output_file}."
        )
    elif STATUS_FAILED in output_file:
        rasa.shared.utils.cli.print_info(
            f"Failing test results have been saved at path: {output_file}."
        )


def transform_results_output_to_yaml(yaml_string: str) -> str:
    """Transform the output of the YAML writer to make it more readable.

    Args:
        yaml_string: The YAML string to transform.

    Returns:
        The transformed YAML string.
    """
    result = []
    for s in yaml_string.splitlines(True):
        if s.startswith("- name"):
            result.append("\n")
            result.append(s)
        elif s.startswith("\n"):
            result.append(s.strip())
        elif s.strip().startswith("#"):
            continue
        else:
            result.append(s)
    return "".join(result)


def _save_tested_commands_histogram(
    count_dict: Dict[str, int], test_status: str, output_dir: str
) -> None:
    """Creates a command histogram and saves it to the specified directory.

    Args:
        count_dict (Dict[str, int]): A dictionary where keys are commands
        and values are counts.
        test_status (str): passing or failing
        output_dir (str): The directory path where the histogram
        image will be saved.
    """
    if not count_dict:
        return

    # Sort the dictionary by keys
    sorted_count_dict = dict(sorted(count_dict.items()))

    plt.figure(figsize=(10, 6))
    bars = plt.bar(sorted_count_dict.keys(), sorted_count_dict.values(), color="blue")
    plt.xlabel("Commands")
    plt.ylabel("Counts")
    plt.title(f"Command histogram for {test_status} tests")
    plt.xticks(rotation=45, ha="right")
    plt.tight_layout()

    # Add total number to each bar
    for bar in bars:
        yval = bar.get_height()
        plt.text(
            bar.get_x() + bar.get_width() / 2,
            yval + 0.5,
            int(yval),
            ha="center",
            va="bottom",
        )

    output_filename = f"commands_histogram_for_{test_status}_tests.png"
    output_file_path = pathlib.Path().joinpath(output_dir, output_filename)
    plt.savefig(output_file_path)
    plt.close()

    structlogger.info(
        "rasa.e2e_test._save_tested_commands_histogram",
        message=f"Commands histogram for {test_status} e2e tests "
        f"are written to '{output_file_path}'.",
    )


def save_test_cases_to_yaml(
    test_results: List["TestResult"],
    output_dir: str,
    status: str,
    test_suite: TestSuite,
) -> None:
    """Extracts TestCases from a list of TestResults and saves them to a YAML file."""
    if not test_results:
        return

    test_cases = [result.test_case for result in test_results]
    new_test_suite = TestSuite(
        test_cases=test_cases,
        fixtures=test_suite.fixtures,
        metadata=test_suite.metadata,
        stub_custom_actions=test_suite.stub_custom_actions,
    )

    output_filename = f"{status}.yml"
    output_file_path = os.path.join(output_dir, output_filename)
    rasa.utils.io.write_yaml(new_test_suite.as_dict(), target=output_file_path)

    structlogger.info(
        "rasa.e2e_test.save_e2e_test_cases",
        message=f"E2e tests with '{status}' status are written to file: "
        f"'{output_file_path}'.",
    )
    if status == STATUS_PASSED:
        structlogger.info(
            "rasa.e2e_test.save_e2e_test_cases",
            message=f"You can use the file: '{output_file_path}' in case you want to "
            f"create training data for fine-tuning an LLM via "
            f"'rasa llm finetune prepare-data'.",
        )



import re
import sys
from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional, Type, Union

import structlog

from rasa.dialogue_understanding.commands import (
    CancelFlowCommand,
    ChitChatAnswerCommand,
    ClarifyCommand,
    Command,
    HumanHandoffCommand,
    KnowledgeAnswerCommand,
    RepeatBotMessagesCommand,
    SetSlotCommand,
    SkipQuestionCommand,
    StartFlowCommand,
)
from rasa.dialogue_understanding.commands.prompt_command import PromptCommand
from rasa.dialogue_understanding.commands.utils import start_flow_by_name
from rasa.shared.core.flows import FlowsList

structlogger = structlog.get_logger()


DEFAULT_COMMANDS = [
    SetSlotCommand,
    StartFlowCommand,
    CancelFlowCommand,
    ChitChatAnswerCommand,
    SkipQuestionCommand,
    KnowledgeAnswerCommand,
    HumanHandoffCommand,
    ClarifyCommand,
    RepeatBotMessagesCommand,
]


@lru_cache(maxsize=128)
def _get_compiled_pattern(pattern_str: str) -> re.Pattern:
    return re.compile(pattern_str)


def _create_default_commands(
    default_commands_to_remove: Union[List[Type[PromptCommand]], List[str]],
) -> List[Type[PromptCommand]]:
    """Return an updated list of default commands after removing the provided commands.

    Args:
        default_commands_to_remove: A list of commands to remove from the default
            commands.

    Returns:
        A copy of the default commands with the provided commands removed.
    """
    commands_after_removal = []
    for command in DEFAULT_COMMANDS:
        if (
            command not in default_commands_to_remove
            and command.__name__ not in default_commands_to_remove
        ):
            commands_after_removal.append(command)

    return commands_after_removal


def _get_additional_parsing_logic(
    command_clz: Type[PromptCommand],
) -> Optional[Callable[[PromptCommand, FlowsList], Optional[PromptCommand]]]:
    """Get additional parsing logic for a command."""
    command_to_parsing_fn_mapper: Dict[
        Type[PromptCommand], Callable[..., Optional[PromptCommand]]
    ] = {
        ClarifyCommand: _parse_clarify_command,
        SetSlotCommand: _parse_set_slot_command,
        StartFlowCommand: _parse_start_flow_command,
    }
    return command_to_parsing_fn_mapper.get(command_clz)


def validate_custom_commands(command_classes: List[Type[PromptCommand]]) -> None:
    clz_not_inheriting_from_command_clz = [
        command_clz
        for command_clz in command_classes
        if not issubclass(command_clz, Command)
    ]

    if clz_not_inheriting_from_command_clz:
        structlogger.error(
            "command_parser.validate_custom_commands.invalid_command",
            invalid_commands=clz_not_inheriting_from_command_clz,
            event_info=(
                "The additional command classes must be a subclass of the 'Command' "
                "class. Please refer to the class in "
                "`rasa.dialogue_understanding.commands.command.Command`"
            ),
        )
        sys.exit(1)

    clz_not_adhering_to_prompt_command_protocol = [
        command_clz
        for command_clz in command_classes
        if not isinstance(command_clz, PromptCommand)
    ]

    if clz_not_adhering_to_prompt_command_protocol:
        structlogger.error(
            "command_parser.validate_custom_commands.invalid_command",
            invalid_commands=clz_not_adhering_to_prompt_command_protocol,
            event_info=(
                "The additional command classes must adhere to the 'PromptCommand' "
                "protocol. Please refer to the protocol in "
                "`rasa.dialogue_understanding.commands.prompt_command.PromptCommand`"
            ),
        )
        sys.exit(1)


def parse_commands(
    actions: Optional[str],
    flows: FlowsList,
    is_handle_flows_prompt: bool = False,
    additional_commands: Optional[List[Type[PromptCommand]]] = None,
    default_commands_to_remove: Optional[
        Union[List[Type[PromptCommand]], List[str]]
    ] = None,
    **kwargs: Any,
) -> List[Command]:
    """Parse a list of action commands."""
    if not actions:
        return []

    commands: List[Command] = []
    validate_custom_commands(additional_commands or [])

    default_commands = DEFAULT_COMMANDS
    if default_commands_to_remove:
        default_commands = _create_default_commands(default_commands_to_remove)

    for action in actions.strip().splitlines():
        if is_handle_flows_prompt:
            # Needed for multistep command generator.
            if (
                len(commands) >= 2
                or len(commands) == 1
                and isinstance(commands[0], ClarifyCommand)
            ):
                break
        commands.extend(
            _parse_standard_commands(default_commands, action.strip(), flows, **kwargs)
        )
        if additional_commands:
            commands.extend(
                _parse_custom_commands(
                    additional_commands, action.strip(), flows, **kwargs
                )
            )
    return commands


def _parse_standard_commands(
    standard_commands: List[Type[PromptCommand]],
    action: str,
    flows: FlowsList,
    **kwargs: Any,
) -> List[Command]:
    """Parse predefined standard commands."""
    commands: List[Command] = []
    for command_clz in standard_commands:
        pattern = _get_compiled_pattern(command_clz.regex_pattern())
        if match := pattern.search(action):
            parsed_command = command_clz.from_dsl(match, **kwargs)
            if _additional_parsing_fn := _get_additional_parsing_logic(command_clz):
                parsed_command = _additional_parsing_fn(parsed_command, flows, **kwargs)
            if parsed_command:
                commands.append(parsed_command)
    return commands


def _parse_custom_commands(
    custom_commands: List[Type[PromptCommand]],
    action: str,
    flows: FlowsList,
    **kwargs: Any,
) -> List[Command]:
    """Parse custom commands."""
    commands: List[Command] = []
    for command_clz in custom_commands:
        pattern = _get_compiled_pattern(command_clz.regex_pattern())
        if match := pattern.search(action):
            if parsed_command := command_clz.from_dsl(match, flows=flows, **kwargs):
                commands.append(parsed_command)
    return commands


def _parse_set_slot_command(
    parsed_command: Optional[SetSlotCommand], flows: FlowsList, **kwargs: Any
) -> Optional[PromptCommand]:
    """Additional parsing logic for the SetSlotCommand."""
    if not parsed_command:
        return None
    # error case where the llm tries to start a flow using a slot set
    if parsed_command.name == "flow_name":
        return start_flow_by_name(parsed_command.value, flows)
    return parsed_command


def _parse_clarify_command(
    parsed_command: Optional[ClarifyCommand], flows: FlowsList, **kwargs: Any
) -> Optional[PromptCommand]:
    """Additional parsing logic for the ClarifyCommand."""
    if not parsed_command:
        return None

    # if no options are available
    if not parsed_command.options:
        # Return the command if options are optional; else, return an empty list
        return (
            ClarifyCommand([])
            if kwargs.get("clarify_options_optional", False)
            else None
        )
    valid_options = [
        flow for flow in parsed_command.options if flow in flows.user_flow_ids
    ]
    unique_valid_options = list(set(valid_options))

    # if there is only one valid option, start the flow
    if len(unique_valid_options) == 1:
        return start_flow_by_name(valid_options[0], flows)

    # if there are multiple valid options, return a clarify command
    if len(valid_options) > 1:
        return ClarifyCommand(valid_options)
    return None


def _parse_start_flow_command(
    parsed_command: Optional[StartFlowCommand], flows: FlowsList, **kwargs: Any
) -> Optional[PromptCommand]:
    """Additional parsing logic for the StartFlowCommand."""
    if not parsed_command:
        return None
    return start_flow_by_name(parsed_command.flow, flows)



from dataclasses import dataclass
from typing import AsyncIterator, Dict, Generic, Optional, Tuple, Type, TypeVar

from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.util import MergeableConfig
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.common import validate_environment


class TTSError(RasaException):
    pass


T = TypeVar("T", bound="TTSEngineConfig")
E = TypeVar("E", bound="TTSEngine")


@dataclass
class TTSEngineConfig(MergeableConfig):
    language: Optional[str] = None
    voice: Optional[str] = None
    timeout: Optional[int] = None


class TTSEngine(Generic[T]):
    required_env_vars: Tuple[str, ...] = ()
    required_packages: Tuple[str, ...] = ()

    def __init__(self, config: Optional[T] = None):
        self.config = self.get_default_config().merge(config)
        validate_environment(
            self.required_env_vars,
            self.required_packages,
            f"TTS Engine {self.__class__.__name__}",
        )

    async def close_connection(self) -> None:
        """Cleanup the connection if necessary."""
        return

    async def synthesize(
        self, text: str, config: Optional[T] = None
    ) -> AsyncIterator[RasaAudioBytes]:
        """Generate speech from text using a remote TTS system."""
        yield RasaAudioBytes(b"")

    def engine_bytes_to_rasa_audio_bytes(self, chunk: bytes) -> RasaAudioBytes:
        """Convert the generated tts audio bytes into rasa audio bytes."""
        raise NotImplementedError

    @staticmethod
    def get_default_config() -> T:
        """Get the default config for this component."""
        raise NotImplementedError

    @classmethod
    def from_config_dict(cls: Type[E], config: Dict) -> E:
        raise NotImplementedError



from typing import TYPE_CHECKING, Any, Dict, Set, Text

from rasa.shared.utils.io import raise_deprecation_warning

if TYPE_CHECKING:
    from rasa.shared.core.flows.flow import FlowLanguageTranslation


RESET_PROPERTY_NAME = "reset_after_flow_ends"
PERSIST_PROPERTY_NAME = "persisted_slots"


def warn_deprecated_collect_step_config() -> None:
    """Warns about deprecated reset_after_flow_ends usage in collect steps."""
    raise_deprecation_warning(
        f"Configuring '{RESET_PROPERTY_NAME}' in collect steps is "
        f"deprecated and will be removed in Rasa Pro 4.0.0. In the parent flow, "
        f"please use the '{PERSIST_PROPERTY_NAME}' "
        "property at the flow level instead."
    )


def get_duplicate_slot_persistence_config_error_message(
    flow_id: str, collect_step: str
) -> str:
    """Returns an error message for duplicate slot persistence configuration."""
    return (
        f"Flow with id '{flow_id}' uses the '{RESET_PROPERTY_NAME}' property "
        f"in collect step '{collect_step}' and also the "
        f"'{PERSIST_PROPERTY_NAME}' property at the flow level. "
        "Please use only one of the two configuration methods."
    )


def get_invalid_slot_persistence_config_error_message(
    flow_id: str, invalid_slots: Set[str]
) -> str:
    """Returns an error message for invalid slot persistence configuration."""
    return (
        f"Flow with id '{flow_id}' lists slot(s) '{invalid_slots}' in the "
        f"'{PERSIST_PROPERTY_NAME}' property. However these slots "
        f"are neither used in a collect step nor a set_slot step of the flow. "
        f"Please remove such slots from the '{PERSIST_PROPERTY_NAME}' property."
    )


def extract_translations(
    translation_data: Dict[Text, Any],
) -> Dict[Text, "FlowLanguageTranslation"]:
    """Extracts translations from a dictionary.

    Args:
        translation_data: The dictionary containing the translations.

    Returns:
        A dictionary containing the extracted translations.
    """
    from rasa.shared.core.flows.flow import FlowLanguageTranslation

    return {
        language_code: FlowLanguageTranslation.parse_obj({**data})
        for language_code, data in translation_data.items()
        if language_code != "metadata"
    }



from __future__ import annotations

from typing import Any, Dict, List, Optional

import structlog

import rasa.dialogue_understanding.stack.utils
from rasa.core.actions.action import Action
from rasa.core.channels import OutputChannel
from rasa.core.nlg import NaturalLanguageGenerator
from rasa.dialogue_understanding.patterns.code_change import FLOW_PATTERN_CODE_CHANGE_ID
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames import (
    BaseFlowStackFrame,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import FlowStackFrameType
from rasa.shared.core.constants import ACTION_CLEAN_STACK
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event
from rasa.shared.core.flows.steps.constants import END_STEP
from rasa.shared.core.flows.steps.continuation import ContinueFlowStep
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


class ActionCleanStack(Action):
    """Action which cancels a flow from the stack."""

    def name(self) -> str:
        """Return the flow name."""
        return ACTION_CLEAN_STACK

    async def run(
        self,
        output_channel: OutputChannel,
        nlg: NaturalLanguageGenerator,
        tracker: DialogueStateTracker,
        domain: Domain,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> List[Event]:
        """Clean the stack."""
        structlogger.debug("action_clean_stack.run")
        new_frames = []
        top_flow_frame = rasa.dialogue_understanding.stack.utils.top_flow_frame(
            tracker.stack, ignore_call_frames=False
        )
        top_user_flow_frame = (
            rasa.dialogue_understanding.stack.utils.top_user_flow_frame(
                tracker.stack, ignore_call_and_link_frames=False
            )
        )

        # Set all frames to their end step, filter out any non-BaseFlowStackFrames
        for frame in tracker.stack.frames:
            if isinstance(frame, BaseFlowStackFrame):
                structlogger.debug(
                    "action_clean_stack.terminating_frame",
                    frame_id=frame.frame_id,
                    flow_id=frame.flow_id,
                )
                frame.step_id = ContinueFlowStep.continue_step_for_id(END_STEP)
                if isinstance(frame, UserFlowStackFrame):
                    # Making sure there are no "continue interrupts" triggered
                    frame.frame_type = FlowStackFrameType.REGULAR
                new_frames.append(frame)
        new_stack = DialogueStack.from_dict([frame.as_dict() for frame in new_frames])

        # Check if the action is being called from within a user flow
        if (
            top_flow_frame
            and top_flow_frame.flow_id != FLOW_PATTERN_CODE_CHANGE_ID
            and top_user_flow_frame
            and top_user_flow_frame.flow_id == top_flow_frame.flow_id
        ):
            # The action is being called from within a user flow on the stack.
            # If there are other frames on the stack, we need to make sure
            # the last executed frame is the end step of the current user flow so
            # that we can trigger pattern_completed for this user flow.
            new_stack.pop()
            structlogger.debug(
                "action_clean_stack.pushing_user_frame_at_the_bottom_of_stack",
                flow_id=top_user_flow_frame.flow_id,
            )
            new_stack.push(
                top_user_flow_frame,
                index=0,
            )

        return tracker.create_stack_updated_events(new_stack)



import os
import shutil
import subprocess
from enum import Enum
from typing import Any, Dict, Optional

import structlog
from pydantic import BaseModel, ConfigDict

from rasa.constants import MODEL_ARCHIVE_EXTENSION
from rasa.model_manager import config
from rasa.model_manager.utils import (
    ensure_base_directory_exists,
    logs_path,
    models_base_path,
    subpath,
    write_encoded_data_to_file,
)
from rasa.model_manager.warm_rasa_process import (
    start_rasa_process,
)
from rasa.model_training import generate_random_model_name

structlogger = structlog.get_logger()


class TrainingSessionStatus(str, Enum):
    """Enum for the training status."""

    RUNNING = "running"
    STOPPED = "stopped"
    DONE = "done"
    ERROR = "error"


class TrainingSession(BaseModel):
    """Store information about a training session."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    training_id: str
    assistant_id: str
    client_id: Optional[str]
    progress: int
    model_name: str
    status: TrainingSessionStatus
    process: subprocess.Popen
    log_id: str

    def is_status_indicating_alive(self) -> bool:
        """Check if the training is running."""
        return self.status == TrainingSessionStatus.RUNNING

    def model_path(self) -> str:
        """Return the path to the model."""
        return subpath(models_base_path(), f"{self.model_name}.tar.gz")


def train_path(training_id: str) -> str:
    """Return the path to the training directory for a given training id."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY + "/trainings", training_id)


def cache_for_assistant_path(assistant_id: str) -> str:
    """Return the path to the cache directory for a given assistant id."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY + "/caches", assistant_id)


def terminate_training(training: TrainingSession) -> None:
    if not training.is_status_indicating_alive():
        # no-op if the training is not running
        return

    structlogger.info(
        "model_trainer.user_stopping_training", training_id=training.training_id
    )
    try:
        training.process.terminate()
        training.status = TrainingSessionStatus.STOPPED
    except ProcessLookupError:
        structlogger.debug(
            "model_trainer.training_process_not_found",
            training_id=training.training_id,
        )
    finally:
        clean_up_after_training(training)


def update_training_status(training: TrainingSession) -> None:
    if not training.is_status_indicating_alive():
        # skip if the training is not running
        return
    if training.process.poll() is None:
        # process is still running
        return

    complete_training(training)


def complete_training(training: TrainingSession) -> None:
    """Complete a training session.

    Transitions the status of a training process to "done" if the process has
    finished successfully, and to "error" if the process has finished with an
    error.
    """
    if training.process.returncode == 0:
        training.status = TrainingSessionStatus.DONE
    else:
        training.status = TrainingSessionStatus.ERROR

    training.progress = 100

    structlogger.info(
        "model_trainer.training_finished",
        training_id=training.training_id,
        status=training.status,
    )

    # persist the assistant cache to speed up future training runs for this
    # assistant
    persist_rasa_cache(training.assistant_id, train_path(training.training_id))
    move_model_to_local_storage(training)

    clean_up_after_training(training)


def clean_up_after_training(training: TrainingSession) -> None:
    """Clean up the training directory."""
    structlogger.debug(
        "model_trainer.cleaning_up_training", training_id=training.training_id
    )
    shutil.rmtree(train_path(training.training_id), ignore_errors=True)


def move_model_to_local_storage(training: TrainingSession) -> None:
    """Persist the model to the remote storage."""
    ensure_base_directory_exists(models_base_path())

    model_path = subpath(
        train_path(training.training_id) + "/models",
        f"{training.model_name}.{MODEL_ARCHIVE_EXTENSION}",
    )

    if os.path.exists(model_path):
        structlogger.debug(
            "model_trainer.persisting_model_to_models_dir",
            training_model_path=model_path,
            storage_model_path=models_base_path(),
        )
        shutil.move(model_path, models_base_path())
    else:
        structlogger.warning(
            "model_trainer.model_not_found_after_training",
            training_id=training.training_id,
            model_path=model_path,
        )


def seed_training_directory_with_rasa_cache(
    training_base_path: str, assistant_id: str
) -> None:
    """Populate the training directory with the cache of a previous training."""
    # check if there is a cache for this assistant
    cache_path = cache_for_assistant_path(assistant_id)

    if os.path.exists(cache_path):
        structlogger.debug(
            "model_trainer.populating_training_dir_with_cache",
            assistant_id=assistant_id,
            training_base_path=training_base_path,
        )
        # copy the cache to the training directory
        shutil.copytree(src=cache_path, dst=subpath(training_base_path, ".rasa"))


def persist_rasa_cache(assistant_id: str, training_base_path: str) -> None:
    """Persist the cache of a training session to speed up future trainings."""
    # copy the cache from the training directory to the cache directory
    # cache files are stored inside of `/.rasa/` of the training folder
    structlogger.debug(
        "model_trainer.persisting_assistant_cache", assistant_id=assistant_id
    )
    cache_path = cache_for_assistant_path(assistant_id)

    # if the training failed and didn't create a cache, skip this step
    if not os.path.exists(subpath(training_base_path, ".rasa")):
        return

    # clean up the cache directory first
    shutil.rmtree(cache_path, ignore_errors=True)
    shutil.copytree(src=subpath(training_base_path, ".rasa"), dst=cache_path)


def write_training_data_to_files(
    encoded_training_data: Dict[str, Any], training_base_path: str
) -> None:
    """Write the training data to files in the training directory.

    Incoming data format, all keys being optional:
    ````
    {
        "domain": "base64 encoded domain.yml",
        "credentials": "base64 encoded credentials.yml",
        "endpoints": "base64 encoded endpoints.yml",
        "flows": "base64 encoded flows.yml",
        "config": "base64 encoded config.yml",
        "stories": "base64 encoded stories.yml",
        "rules": "base64 encoded rules.yml",
        "nlu": "base64 encoded nlu.yml"
    }
    ```
    """
    data_to_be_written_to_files = {
        "domain": "domain.yml",
        "credentials": "credentials.yml",
        "endpoints": "endpoints.yml",
        "flows": "data/flows.yml",
        "config": "config.yml",
        "stories": "data/stories.yml",
        "rules": "data/rules.yml",
        "nlu": "data/nlu.yml",
    }

    for key, file in data_to_be_written_to_files.items():
        parent_path, file_name = os.path.split(file)

        write_encoded_data_to_file(
            encoded_training_data.get(key, ""),
            subpath(training_base_path + "/" + parent_path, file_name),
        )


def prepare_training_directory(
    training_base_path: str, assistant_id: str, encoded_training_data: Dict[str, Any]
) -> None:
    """Prepare the training directory for a new training session."""
    # create a new working directory and store the training data from the
    # request there. the training data in the request is base64 encoded
    os.makedirs(training_base_path, exist_ok=True)

    seed_training_directory_with_rasa_cache(training_base_path, assistant_id)
    write_training_data_to_files(encoded_training_data, training_base_path)
    structlogger.debug("model_trainer.prepared_training", path=training_base_path)


def start_training_process(
    training_id: str,
    assistant_id: str,
    client_id: str,
    training_base_path: str,
) -> TrainingSession:
    model_name = generate_random_model_name()
    # Start the training in a subprocess
    # set the working directory to the training directory
    # run the rasa train command as a subprocess, activating poetry before running
    # pipe the stdout and stderr to the same file
    arguments = [
        "train",
        "--debug",
        "--data",
        "data",
        "--config",
        "config.yml",
        "--domain",
        "domain.yml",
        "--endpoints",
        "endpoints.yml",
        "--fixed-model-name",
        f"{model_name}.{MODEL_ARCHIVE_EXTENSION}",
        "--out",
        "models",
    ]

    if config.SERVER_MODEL_REMOTE_STORAGE:
        arguments.extend(
            [
                "--keep-local-model-copy",
                "--remote-storage",
                config.SERVER_MODEL_REMOTE_STORAGE,
                "--remote-root-only",
            ]
        )

    structlogger.debug(
        "model_trainer.training_arguments", arguments=" ".join(arguments)
    )

    warm_process = start_rasa_process(cwd=training_base_path, arguments=arguments)

    structlogger.info(
        "model_trainer.training_started",
        training_id=training_id,
        assistant_id=assistant_id,
        model_name=model_name,
        client_id=client_id,
        log=logs_path(warm_process.log_id),
        pid=warm_process.process.pid,
    )

    return TrainingSession(
        training_id=training_id,
        assistant_id=assistant_id,
        client_id=client_id,
        model_name=model_name,
        progress=0,
        status=TrainingSessionStatus.RUNNING,
        process=warm_process.process,  # Store the process handle
        log_id=warm_process.log_id,
    )


def run_training(
    training_id: str,
    assistant_id: str,
    client_id: str,
    encoded_training_data: Dict,
) -> TrainingSession:
    """Run a training session."""
    training_base_path = train_path(training_id)

    prepare_training_directory(training_base_path, assistant_id, encoded_training_data)
    return start_training_process(
        training_id=training_id,
        assistant_id=assistant_id,
        client_id=client_id,
        training_base_path=training_base_path,
    )



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import Any, Dict, Optional

import structlog

from rasa.shared.constants import (
    API_BASE_CONFIG_KEY,
    API_TYPE_CONFIG_KEY,
    API_VERSION_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    N_REPHRASES_CONFIG_KEY,
    OPENAI_API_BASE_CONFIG_KEY,
    OPENAI_API_TYPE_CONFIG_KEY,
    OPENAI_API_VERSION_CONFIG_KEY,
    OPENAI_PROVIDER,
    PROVIDER_CONFIG_KEY,
    RASA_TYPE_CONFIG_KEY,
    REQUEST_TIMEOUT_CONFIG_KEY,
    SELF_HOSTED_PROVIDER,
    STREAM_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
    USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY,
)
from rasa.shared.providers._configs.utils import (
    raise_deprecation_warnings,
    resolve_aliases,
    validate_forbidden_keys,
    validate_required_keys,
)

structlogger = structlog.get_logger()


DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING = {
    # Model name aliases
    MODEL_NAME_CONFIG_KEY: MODEL_CONFIG_KEY,
    # Provider aliases
    RASA_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    # API type aliases
    OPENAI_API_TYPE_CONFIG_KEY: API_TYPE_CONFIG_KEY,
    # API base aliases
    OPENAI_API_BASE_CONFIG_KEY: API_BASE_CONFIG_KEY,
    # API version aliases
    OPENAI_API_VERSION_CONFIG_KEY: API_VERSION_CONFIG_KEY,
    # Timeout aliases
    REQUEST_TIMEOUT_CONFIG_KEY: TIMEOUT_CONFIG_KEY,
}

REQUIRED_KEYS = [API_BASE_CONFIG_KEY, MODEL_CONFIG_KEY, PROVIDER_CONFIG_KEY]

FORBIDDEN_KEYS = [
    STREAM_CONFIG_KEY,
    N_REPHRASES_CONFIG_KEY,
]


@dataclass
class SelfHostedLLMClientConfig:
    """Parses configuration for Self Hosted LiteLLM client.

    Resolves aliases and raises deprecation warnings.

    Raises:
        ValueError: Raised in cases of invalid configuration:
            - If any of the required configuration keys are missing.
    """

    model: str
    provider: str
    api_base: str
    api_version: Optional[str] = None
    api_type: Optional[str] = OPENAI_PROVIDER
    use_chat_completions_endpoint: Optional[bool] = True
    extra_parameters: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        if self.model is None:
            message = "Model cannot be set to None."
            structlogger.error(
                "self_hosted_llm_client_config.validation_error",
                message=message,
                model=self.model,
            )
            raise ValueError(message)
        if self.provider is None:
            message = "Provider cannot be set to None."
            structlogger.error(
                "self_hosted_llm_client_config.validation_error",
                message=message,
                provider=self.provider,
            )
            raise ValueError(message)
        if self.api_base is None:
            message = "API base cannot be set to None."
            structlogger.error(
                "self_hosted_llm_client_config.validation_error",
                message=message,
                provider=self.provider,
            )
            raise ValueError(message)
        if self.api_type != OPENAI_PROVIDER:
            message = (
                f"Currently supports only {OPENAI_PROVIDER} endpoints. "
                f"API type must be set to '{OPENAI_PROVIDER}'."
            )
            structlogger.error(
                "self_hosted_llm_client_config.validation_error",
                message=message,
                api_type=self.api_type,
            )
            raise ValueError(message)

    @classmethod
    def from_dict(cls, config: dict) -> SelfHostedLLMClientConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Config is missing required keys.

        Returns:
            DefaultLiteLLMClientConfig
        """
        # Check for deprecated keys
        raise_deprecation_warnings(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)
        # Resolve any potential aliases
        config = cls.resolve_config_aliases(config)
        # Validate that the required keys are present
        validate_required_keys(config, REQUIRED_KEYS)
        # Validate that the forbidden keys are not present
        validate_forbidden_keys(config, FORBIDDEN_KEYS)
        this = SelfHostedLLMClientConfig(
            # Required parameters
            model=config.pop(MODEL_CONFIG_KEY),
            provider=config.pop(PROVIDER_CONFIG_KEY),
            api_base=config.pop(API_BASE_CONFIG_KEY),
            # Optional parameters
            api_type=config.pop(API_TYPE_CONFIG_KEY, OPENAI_PROVIDER),
            api_version=config.pop(API_VERSION_CONFIG_KEY, None),
            use_chat_completions_endpoint=config.pop(
                USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY, True
            ),
            # The rest of parameters (e.g. model parameters) are considered
            # as extra parameters
            extra_parameters=config,
        )
        return this

    def to_dict(self) -> dict:
        """Converts the config instance into a dictionary."""
        d = asdict(self)
        # Extra parameters should also be on the top level
        d.pop("extra_parameters", None)
        d.update(self.extra_parameters)
        return d

    @staticmethod
    def resolve_config_aliases(config: Dict[str, Any]) -> Dict[str, Any]:
        return resolve_aliases(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)


def is_self_hosted_config(config: dict) -> bool:
    """Check whether the configuration is meant to configure an self-hosted client."""
    # Process the config to handle all the aliases
    config = SelfHostedLLMClientConfig.resolve_config_aliases(config)

    # Case: Configuration contains `provider: self-hosted`
    if config.get(PROVIDER_CONFIG_KEY) == SELF_HOSTED_PROVIDER:
        return True

    return False



DEFAULT_AZURE_API_KEY_NAME = "AZURE_API_KEY"
LITE_LLM_API_BASE_FIELD = "api_base"
LITE_LLM_API_KEY_FIELD = "api_key"
LITE_LLM_API_VERSION_FIELD = "api_version"
LITE_LLM_MODEL_FIELD = "model"
LITE_LLM_AZURE_AD_TOKEN = "azure_ad_token"



import asyncio
import copy
import datetime
import difflib
from asyncio import CancelledError
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, DefaultDict, Dict, List, Optional, Text, Tuple, Union
from urllib.parse import urlparse

import requests
import structlog
from tqdm import tqdm

import rasa.shared.utils.io
from rasa.core.channels import CollectingOutputChannel, UserMessage
from rasa.core.constants import ACTIVE_FLOW_METADATA_KEY, STEP_ID_METADATA_KEY
from rasa.core.exceptions import AgentNotReady
from rasa.core.persistor import StorageType
from rasa.core.utils import AvailableEndpoints
from rasa.dialogue_understanding_test.du_test_case import DialogueUnderstandingTestCase
from rasa.e2e_test.constants import TEST_CASE_NAME, TEST_FILE_NAME
from rasa.e2e_test.e2e_config import create_llm_judge_config
from rasa.e2e_test.e2e_test_case import (
    KEY_STUB_CUSTOM_ACTIONS,
    ActualStepOutput,
    Fixture,
    Metadata,
    TestCase,
    TestStep,
)
from rasa.e2e_test.e2e_test_result import (
    NO_RESPONSE,
    NO_SLOT,
    TestFailure,
    TestResult,
)
from rasa.llm_fine_tuning.conversations import Conversation
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX
from rasa.shared.core.events import (
    ActionExecuted,
    BotUttered,
    Event,
    FlowCompleted,
    FlowStarted,
    SlotSet,
    UserUttered,
)
from rasa.shared.core.flows.flow_path import FlowPath, PathNode
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import COMMANDS
from rasa.telemetry import track_e2e_test_run
from rasa.utils.endpoints import EndpointConfig

structlogger = structlog.get_logger()

TEST_TURNS_TYPE = Dict[int, Union[TestStep, ActualStepOutput]]


class E2ETestRunner:
    def __init__(
        self,
        model_path: Optional[Text] = None,
        model_server: Optional[EndpointConfig] = None,
        remote_storage: Optional[StorageType] = None,
        endpoints: Optional[AvailableEndpoints] = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the E2E test suite runner.

        Args:
            model_path: Path to the model.
            model_server: Model server configuration.
            remote_storage: Remote storage to use for model retrieval.
            endpoints: Endpoints configuration.
            **kwargs: Additional arguments
        """
        import rasa.core.agent

        structlogger.info(
            "e2e_test_runner.init",
            event_info="Started running end-to-end testing.",
        )

        test_case_path = kwargs.get("test_case_path")
        self.llm_judge_config = create_llm_judge_config(test_case_path)

        are_custom_actions_stubbed = (
            endpoints
            and endpoints.action
            and endpoints.action.kwargs.get(KEY_STUB_CUSTOM_ACTIONS)
        )
        if endpoints and not are_custom_actions_stubbed:
            self._action_server_is_reachable(endpoints)

        self.agent = asyncio.run(
            rasa.core.agent.load_agent(
                model_path=model_path,
                model_server=model_server,
                remote_storage=remote_storage,
                endpoints=endpoints,
            )
        )

        if not self.agent.is_ready():
            raise AgentNotReady(
                "Agent needs to be prepared before usage. "
                "Please check that the agent was able to "
                "load the trained model."
            )

    async def run_prediction_loop(
        self,
        collector: CollectingOutputChannel,
        steps: List[TestStep],
        sender_id: Text,
        test_case_metadata: Optional[Metadata] = None,
        input_metadata: Optional[List[Metadata]] = None,
    ) -> TEST_TURNS_TYPE:
        """Runs dialogue prediction.

        Args:
            collector: Output channel.
            steps: List of steps to run.
            sender_id: The test case name with added timestamp suffix.
            test_case_metadata: Metadata of the test case.
            input_metadata: List of metadata.

        Returns:
            Test turns: {turn_sequence (int): TestStep or ActualStepOutput}.
        """
        turns: TEST_TURNS_TYPE = {}
        event_cursor = 0

        if not self.agent.processor:
            return turns

        tracker = await self.agent.processor.fetch_tracker_with_initial_session(
            sender_id, output_channel=collector
        )
        # turn -1 i used to contain events that happen during
        # the start of the session and before the first user message
        # TestStep is a placeholder just for the sake of having a turn
        # to specify the actor
        turns[-1], event_cursor = self.get_actual_step_output(
            tracker,
            TestStep(
                actor="bot",
                text=None,
            ),
            event_cursor,
        )

        # This variable tracks the position of the last user step.
        # We use it to determine from which turn onward we should check
        # for failures in case of the fail-fast logic.
        last_user_step_position = 0
        for position, step in enumerate(steps):
            if step.actor != "user":
                turns[position] = step

                # Check for failures after each bot step
                try:
                    self.fail_fast_check(
                        test_turns=turns,
                        last_user_step_position=last_user_step_position,
                    )
                except RasaException:
                    break

                continue
            elif not step.text:
                rasa.shared.utils.io.raise_warning(
                    f"The test case '{sender_id}' contains a `user` step in line "
                    f"{position + 1} without a text value. "
                    f"Skipping this step and proceeding to the next user step.",
                    UserWarning,
                )
                continue
            # Update to the current position because we have a valid user step
            last_user_step_position = position

            metadata = test_case_metadata.metadata if test_case_metadata else {}

            if input_metadata:
                step_metadata = self.filter_metadata_for_input(
                    step.metadata_name, input_metadata
                )
                step_metadata_dict = step_metadata.metadata if step_metadata else {}
                metadata = self.merge_metadata(
                    sender_id, step.text, metadata, step_metadata_dict
                )

            try:
                await self.agent.handle_message(
                    UserMessage(
                        step.text,
                        collector,
                        sender_id,
                        metadata=metadata,
                    )
                )
            except CancelledError:
                structlogger.error(
                    "e2e_test_runner.run_prediction_loop",
                    error=f"Message handling timed out for user message '{step.text}'.",
                    exc_info=True,
                )
            except Exception as exc:
                structlogger.error(
                    "e2e_test_runner.run_prediction_loop",
                    error=f"An exception occurred while handling "
                    f"user message '{step.text}'. Error: {exc}",
                )
            tracker = await self.agent.tracker_store.retrieve(sender_id)  # type: ignore[assignment]
            turns[position], event_cursor = self.get_actual_step_output(
                tracker, step, event_cursor
            )
        return turns

    def fail_fast_check(
        self, test_turns: TEST_TURNS_TYPE, last_user_step_position: int
    ) -> None:
        """Checks whether there are any test failures in 'test_turns'.

        Args:
            test_turns (TEST_TURNS_TYPE): The transcript of test cases and events.
            last_user_step_position (int): The start position in the test turns.

        Raises:
            RasaException if a failure is found.
        """
        failures_found = self.find_test_failures(
            test_turns=test_turns,
            test_case=None,
            last_user_step_position=last_user_step_position,
        )
        if failures_found:
            raise RasaException("Test failure found. Aborting early.")

    @staticmethod
    def merge_metadata(
        sender_id: Text,
        step_text: Text,
        test_case_metadata: Dict[Text, Text],
        step_metadata: Dict[Text, Text],
    ) -> Dict[Text, Text]:
        """Merges the test case and user step metadata.

        Args:
            sender_id: The test case name with added timestamp suffix.
            step_text: The user step text.
            test_case_metadata: The test case metadata dict.
            step_metadata: The user step metadata dict.

        Returns:
            A dictionary with the merged metadata.
        """
        if not test_case_metadata:
            return step_metadata
        if not step_metadata:
            return test_case_metadata

        keys_to_overwrite = []

        for key in step_metadata.keys():
            if key in test_case_metadata.keys():
                keys_to_overwrite.append(key)

        if keys_to_overwrite:
            test_case_name = sender_id.rsplit("_", 1)[0]
            structlogger.warning(
                "e2e_test_runner.merge_metadata",
                message=f"Metadata {keys_to_overwrite} exist in both the test case "
                f"'{test_case_name}' and the user step '{step_text}'. "
                "The user step metadata takes precedence and will "
                "override the test case metadata.",
            )

        merged_metadata = copy.deepcopy(test_case_metadata)
        merged_metadata.update(step_metadata)

        return merged_metadata

    @staticmethod
    def get_actual_step_output(
        tracker: DialogueStateTracker,
        test_step: TestStep,
        event_cursor: int,
    ) -> Tuple[ActualStepOutput, int]:
        """Returns the events that are generated from the current step.

        Args:
            tracker: The tracker for the current test case.
            test_step: The test step.
            event_cursor: The event cursor where the previous step left off.

        Returns:
        The events generated from the current step and the updated event
        cursor position.
        """
        session_events = list(tracker.events)[event_cursor:]
        if session_events:
            event_cursor += len(session_events)
            return (
                ActualStepOutput.from_test_step(
                    test_step,
                    [
                        event
                        for event in session_events
                        if isinstance(event, (BotUttered, UserUttered, SlotSet))
                    ],
                ),
                event_cursor,
            )
        else:
            structlogger.warning(
                "e2e_test_runner.get_actual_step_output",
                message=f"No events found for '{tracker.sender_id}' after processing "
                f"test step '{test_step.text}'.",
            )
            # if there are no events, we still want to return an
            # ActualStepOutput object with the test step as the
            # response to be able to generate a diff
            return (
                ActualStepOutput.from_test_step(
                    test_step,
                    [
                        UserUttered(text=test_step.text),
                        BotUttered(text=NO_RESPONSE),
                    ],
                ),
                event_cursor,
            )

    @classmethod
    def generate_test_result(
        cls,
        test_turns: TEST_TURNS_TYPE,
        test_case: TestCase,
    ) -> TestResult:
        """Generates test result.

        Args:
            test_turns: the turns that happened when running the test case or test step.
            test_case: the `TestCase` instance.

        Returns:
        Test result.
        """
        difference = []
        error_line = None
        test_failures = cls.find_test_failures(test_turns, test_case)
        if test_failures:
            first_failure = test_failures[0][0]
            difference = cls.human_readable_diff(test_turns, test_failures)
            error_line = first_failure.error_line if first_failure else None

        return TestResult(
            pass_status=len(test_failures) == 0,
            test_case=test_case,
            difference=difference,
            error_line=error_line,
        )

    def _get_additional_splitting_conditions(
        self,
        step: TestStep,
        input_metadata: List[Metadata],
        tracker: DialogueStateTracker,
        test_case: TestCase,
    ) -> Dict[str, Any]:
        """Returns additional splitting conditions for the user message."""
        additional_splitting_conditions: Dict[str, Any] = {"text": step.text}

        if not step.metadata_name:
            return additional_splitting_conditions

        step_metadata = self.filter_metadata_for_input(
            step.metadata_name, input_metadata
        )
        step_metadata_dict = step_metadata.metadata if step_metadata else {}

        test_case_metadata = self.filter_metadata_for_input(
            test_case.metadata_name, input_metadata
        )
        test_case_metadata_as_dict = (
            test_case_metadata.metadata if test_case_metadata else {}
        )

        metadata: Dict[str, Any] = self.merge_metadata(
            tracker.sender_id,
            step.text,
            test_case_metadata_as_dict,
            step_metadata_dict,
        )
        metadata["model_id"] = tracker.model_id
        metadata["assistant_id"] = tracker.assistant_id

        additional_splitting_conditions["metadata"] = metadata

        return additional_splitting_conditions

    @staticmethod
    def _get_current_user_turn_and_prior_events(
        tracker: DialogueStateTracker,
        additional_splitting_conditions: Dict[str, Any],
        step: TestStep,
    ) -> Tuple[List[Event], List[Event]]:
        """Returns the current user turn and prior events."""
        actual_events = tracker.events

        # this returns 2 lists, the first list contains the events until the user
        # message and the second list contains the events after the
        # user message, including the user message
        step_events = rasa.shared.core.events.split_events(
            actual_events,
            UserUttered,
            additional_splitting_conditions=additional_splitting_conditions,
            include_splitting_event=True,
        )

        if len(step_events) < 2:
            structlogger.error(
                "e2e_test_runner.run_assertions.user_message_not_found",
                message=f"User message '{step.text}' was not found in "
                f"the actual events. The user message "
                f"properties which were searched: "
                f"{additional_splitting_conditions}",
            )
            return [], []

        post_step_events = step_events[1]
        prior_events = step_events[0]

        # subset of events until the next user message
        turn_events = []
        for event in post_step_events:
            # we reached the next user message
            if isinstance(event, UserUttered) and step.text != event.text:
                break

            turn_events.append(event)

        return turn_events, prior_events

    @staticmethod
    def _slice_turn_events(
        step: TestStep,
        matching_event: Event,
        turn_events: List[Event],
        prior_events: List[Event],
    ) -> Tuple[List[Event], List[Event]]:
        """Slices the turn events when assertion order is enabled."""
        if not step.assertion_order_enabled:
            return turn_events, prior_events

        if not matching_event:
            return turn_events, prior_events

        matching_event_index = turn_events.index(matching_event)
        if matching_event_index + 1 < len(turn_events):
            prior_events += turn_events[: matching_event_index + 1]
            turn_events = turn_events[matching_event_index + 1 :]

        return turn_events, prior_events

    async def run_assertions(
        self,
        sender_id: str,
        test_case: TestCase,
        input_metadata: Optional[List[Metadata]],
    ) -> TestResult:
        """Runs the assertions defined in the test case."""
        tracker = await self.agent.processor.get_tracker(sender_id)  # type: ignore[union-attr]

        assertion_failure = None
        assertion_failure_found = False
        input_metadata = input_metadata if input_metadata else []

        for index, step in enumerate(test_case.steps):
            if not step.assertions:
                structlogger.debug(
                    "e2e_test_runner.run_assertions.no_assertions.skipping_step",
                    step=step,
                )
                continue

            additional_splitting_conditions = self._get_additional_splitting_conditions(
                step, input_metadata, tracker, test_case
            )

            turn_events, prior_events = self._get_current_user_turn_and_prior_events(
                tracker, additional_splitting_conditions, step
            )

            if not turn_events:
                return TestResult(
                    pass_status=False,
                    test_case=test_case,
                    difference=[],
                    error_line=step.line,
                    assertion_failure=None,
                )

            for assertion in step.assertions:
                structlogger.debug(
                    "e2e_test_runner.run_assertions.running_assertion",
                    test_case_name=test_case.name,
                    step_text=step.text,
                    assertion_type=assertion.type(),
                )

                assertion_order_error_msg = ""

                if step.assertion_order_enabled:
                    assertion_order_error_msg = (
                        " You have enabled assertion order, "
                        "you should check the order in which the "
                        "assertions are listed for this user step."
                    )

                assertion_failure, matching_event = assertion.run(
                    turn_events,
                    prior_events=prior_events,
                    assertion_order_error_message=assertion_order_error_msg,
                    llm_judge_config=self.llm_judge_config,
                    step_text=step.text,
                    step_index=index,
                )

                if assertion_failure:
                    assertion_failure_found = True
                    structlogger.debug(
                        "e2e_test_runner.run_assertions.assertion_failure_found",
                        test_case_name=test_case.name,
                        error_line=assertion_failure.error_line,
                    )
                    break

                turn_events, prior_events = self._slice_turn_events(
                    step, matching_event, turn_events, copy.deepcopy(prior_events)
                )

            if assertion_failure_found:
                # don't continue with the next steps if an assertion failed
                break

        return TestResult(
            pass_status=not assertion_failure,
            test_case=test_case,
            difference=[],
            error_line=assertion_failure.error_line if assertion_failure else None,
            assertion_failure=assertion_failure,
        )

    @classmethod
    def _resolve_successful_bot_utter(
        cls,
        expected_result: TestStep,
        test_response: ActualStepOutput,
    ) -> str:
        """Returns the diff text for a successful bot utter test step."""
        for event in test_response.bot_uttered_events:
            if expected_result.matches_event(event):
                # remove the event that is already matched so
                # that we dont compare against it again
                test_response.remove_bot_uttered_event(event)
                if expected_result.text is not None:
                    text = f"{expected_result.actor}: {expected_result.text}"
                elif expected_result.template is not None:
                    text = f"{expected_result.actor}: {expected_result.template}"
                break

        return text

    @classmethod
    def _resolve_successful_set_slot(
        cls,
        expected_result: TestStep,
        test_response: ActualStepOutput,
    ) -> str:
        """Returns the diff text for a successful set slot test step."""
        slot_name = expected_result.get_slot_name()
        text = f"slot_was_set: {slot_name}"

        for event in test_response.slot_set_events:
            if expected_result.matches_event(event):
                # remove the event that is already matched so
                # that we dont compare against it again
                test_response.remove_slot_set_event(event)
                if event.value is not None:
                    text += f": {event.value} ({type(event.value).__name__})"
                break

        return text

    @classmethod
    def _resolve_successful_slot_was_not_set(cls, expected_result: TestStep) -> str:
        """Returns the diff text for a successful slot was not set test step."""
        slot_name = expected_result.get_slot_name()
        slot_value = expected_result.get_slot_value()
        text = f"slot_was_not_set: {slot_name}"
        if expected_result.is_slot_instance_dict():
            text += f": {slot_value}"

        return text

    @classmethod
    def _find_first_non_matched_utterance(
        cls,
        test_response: ActualStepOutput,
        bot_utter_test_steps: List[TestStep],
    ) -> Optional[BotUttered]:
        """Finds the first non matched utterance in events (if any)."""
        bot_events = test_response.bot_uttered_events
        matching_events = []
        for event in bot_events:
            for test_step in bot_utter_test_steps:
                if test_step.matches_event(event):
                    matching_events.append(event)
                    break
        not_matching_events = [
            event for event in bot_events if event not in matching_events
        ]
        return not_matching_events[0] if not_matching_events else None

    @classmethod
    def _handle_fail_diff(
        cls,
        failed_step: TestStep,
        test_response: ActualStepOutput,
        bot_utter_test_steps: List[TestStep],
    ) -> Tuple[str, str]:
        """Handles generating the diff text for a failed test step."""
        if failed_step.text is not None:
            diff_test_text = f"{failed_step.actor}: {failed_step.text}"
            diff_actual_text = NO_RESPONSE
            event: Optional[Union[BotUttered, SlotSet]] = (
                cls._find_first_non_matched_utterance(
                    test_response, bot_utter_test_steps
                )
            )
            if event and isinstance(event, BotUttered):
                test_response.remove_bot_uttered_event(event)
                diff_actual_text = f"bot: {event.text}"

        elif failed_step.template is not None:
            diff_test_text = f"{failed_step.actor}: {failed_step.template}"
            diff_actual_text = NO_RESPONSE
            event = cls._find_first_non_matched_utterance(
                test_response, bot_utter_test_steps
            )
            if event:
                test_response.remove_bot_uttered_event(event)
                diff_actual_text = f"bot: {event.metadata.get('utter_action')}"

        elif failed_step.slot_was_set:
            slot_name = failed_step.get_slot_name()

            diff_test_text = f"slot_was_set: {slot_name}"
            if failed_step.is_slot_instance_dict():
                slot_value = failed_step.get_slot_value()
                diff_test_text += f": {slot_value} ({type(slot_value).__name__})"

            diff_actual_text = NO_SLOT
            for event in test_response.slot_set_events:
                if slot_name == event.key:
                    diff_actual_text = (
                        f"slot_was_set: {event.key}: {event.value} "
                        f"({type(event.value).__name__})"
                    )
                    test_response.remove_slot_set_event(event)
                    break

        elif failed_step.slot_was_not_set:
            slot_name = failed_step.get_slot_name()

            diff_test_text = f"slot_was_not_set: {slot_name}"
            if failed_step.is_slot_instance_dict():
                slot_value = failed_step.get_slot_value()
                diff_test_text += f": {slot_value}"

            for event in test_response.slot_set_events:
                if slot_name == event.key:
                    diff_actual_text = (
                        f"slot_was_set: {event.key}: {event.value} "
                        f"({type(event.value).__name__})"
                    )
                    test_response.remove_slot_set_event(event)
                    break

        return diff_test_text, diff_actual_text

    @classmethod
    def _select_bot_utter_turns(
        cls,
        test_turns: TEST_TURNS_TYPE,
        start_index: int,
    ) -> List[TestStep]:
        """Selects the TestSteps from the test turns that match BotUttered events."""
        bot_utter_turns = []
        for index in range(start_index, len(test_turns) - 1):
            test_turn = test_turns[index]
            if isinstance(test_turn, TestStep):
                if test_turn.text is not None or test_turn.template is not None:
                    bot_utter_turns.append(test_turn)
            elif isinstance(test_turn, ActualStepOutput):
                break

        return bot_utter_turns

    @classmethod
    def human_readable_diff(
        cls,
        test_turns: TEST_TURNS_TYPE,
        fail_positions: List[Tuple[TestFailure, int]],
    ) -> List[str]:
        """Returns a human readable diff of the test case and the actual conversation.

        Given an ordered list of test steps and actual conversation events, this
        method will return a human readable diff of the two.
        The diff uses difflib to compare the two lists and will highlight
        differences between the two in a pytest style diff.

        Args:
            test_turns: The transcript of test cases and events.
            fail_positions: The positions of the test failures.

        Returns:
            The human readable diff.
        """
        actual_transcript = []
        expected_transcript = []
        failure_points = {position for _, position in fail_positions}
        # This will only be used when the TestCase is not started
        # with a user step
        latest_response: ActualStepOutput = test_turns[-1]  # type: ignore[assignment]

        for index in range(max(failure_points) + 1):
            test_result = test_turns[index]
            if index in failure_points:
                diff_test_text, diff_actual_text = cls._handle_fail_diff(
                    test_result,  # type: ignore[arg-type]
                    latest_response,
                    cls._select_bot_utter_turns(test_turns, index),
                )  # test_result can only be TestStep in failure_points
                actual_transcript.append(diff_actual_text)
                expected_transcript.append(diff_test_text)
                continue

            if isinstance(test_result, TestStep):
                if test_result.text is not None or test_result.template is not None:
                    diff_test_text = cls._resolve_successful_bot_utter(
                        expected_result=test_result,
                        test_response=latest_response,
                    )

                if test_result.slot_was_set:
                    diff_test_text = cls._resolve_successful_set_slot(
                        expected_result=test_result,
                        test_response=latest_response,
                    )

                if test_result.slot_was_not_set:
                    diff_test_text = cls._resolve_successful_slot_was_not_set(
                        expected_result=test_result,
                    )

            elif isinstance(test_result, ActualStepOutput):
                latest_response = test_result
                event = test_result.get_user_uttered_event()
                if event:
                    diff_test_text = f"{test_result.actor}: {event.text}"
                else:
                    raise RasaException(
                        f"Bot did not catch user "
                        f"event: {test_result.actor}: {test_result.text}."
                    )
            else:
                raise ValueError(f"Unexpected test result type: {type(test_result)}")

            # test passed in these cases so it's the same
            actual_transcript.append(diff_test_text)
            expected_transcript.append(diff_test_text)

        return list(difflib.ndiff(actual_transcript, expected_transcript))

    @classmethod
    def find_test_failures(
        cls,
        test_turns: TEST_TURNS_TYPE,
        test_case: Optional[TestCase],
        last_user_step_position: int = 0,
    ) -> List[Tuple[TestFailure, int]]:
        """Finds the test failures in the transcript.

        Args:
            test_turns (TEST_TURNS_TYPE): The transcript of test cases and events.
            test_case (Optional[TestCase]): The test case.
            last_user_step_position (int): The start position in the test turns.

        Returns:
            The test failures or an empty list if there is no test failure.
        """
        # This will only be used when the TestCase is not started
        # with a user step
        latest_response: ActualStepOutput = test_turns[-1]  # type: ignore[assignment]
        failures = []
        match = None
        for position in range(last_user_step_position, len(test_turns) - 1):
            if position not in test_turns:
                continue
            turn_value = test_turns[position]
            if isinstance(turn_value, ActualStepOutput):
                latest_response = turn_value
            elif isinstance(turn_value, TestStep):
                if turn_value.text is not None or turn_value.template is not None:
                    match = cls._does_match_exist(
                        latest_response.bot_uttered_events, turn_value
                    )
                if turn_value.slot_was_set:
                    match = cls._does_match_exist(
                        latest_response.slot_set_events, turn_value
                    )
                if turn_value.slot_was_not_set:
                    match = not cls._does_match_exist(
                        latest_response.slot_set_events, turn_value
                    )
                if not match:
                    failures.append((TestFailure(test_case, turn_value.line), position))
            else:
                raise ValueError(f"Unexpected turn value type: {type(turn_value)}")

        return failures

    @classmethod
    def _does_match_exist(
        cls,
        actual_events: Optional[List[Union[UserUttered, BotUttered, SlotSet]]],
        expected: TestStep,
    ) -> bool:
        if not actual_events:
            return False

        match = False
        for event in actual_events:
            if expected.matches_event(event):
                return True
        return match

    async def set_up_fixtures(
        self,
        fixtures: List[Fixture],
        sender_id: Text,
    ) -> None:
        """Sets slots in the tracker as defined by the input fixtures.

        Args:
            fixtures: List of `Fixture` objects.
            sender_id: The conversation id.
        """
        if not fixtures:
            return
        if not self.agent.processor:
            return

        tracker = await self.agent.processor.fetch_tracker_with_initial_session(
            sender_id, output_channel=CollectingOutputChannel()
        )

        for fixture in fixtures:
            for slot_name, slot_value in fixture.slots_set.items():
                tracker.update(SlotSet(slot_name, slot_value))

        await self.agent.tracker_store.save(tracker)

    @staticmethod
    def filter_fixtures_for_test_case(
        test_case: TestCase, fixtures: List[Fixture]
    ) -> List[Fixture]:
        """Filters the input fixtures for the input test case.

        Args:
            test_case: The test case.
            fixtures: The fixtures.

        Returns:
        The filtered fixtures.
        """
        return list(
            filter(
                lambda fixture: test_case.fixture_names
                and fixture.name in test_case.fixture_names,
                fixtures,
            )
        )

    @staticmethod
    def filter_metadata_for_input(
        metadata_name: Optional[Text], test_suite_metadata: List[Metadata]
    ) -> Optional[Metadata]:
        """Filters the test suite metadata for a metadata name.

        Args:
            metadata_name: The test case or user step metadata name.
            test_suite_metadata: The top level list of all metadata definitions.

        Returns:
            The filtered metadata.
        """
        if not metadata_name:
            return None

        filtered_metadata = list(
            filter(
                lambda metadata: metadata_name and metadata.name == metadata_name,
                test_suite_metadata,
            )
        )

        if not filtered_metadata:
            structlogger.warning(
                "e2e_test_runner.filter_metadata_for_input",
                message=f"Metadata '{metadata_name}' is not defined in the input "
                f"metadata.",
            )
            return None

        return filtered_metadata[0]

    async def run_tests(
        self,
        input_test_cases: List[TestCase],
        input_fixtures: List[Fixture],
        fail_fast: bool = False,
        **kwargs: Any,
    ) -> List["TestResult"]:
        """Runs the test cases.

        Args:
            input_test_cases: Input test cases.
            input_fixtures: Input fixtures.
            fail_fast: Whether to fail fast.
            **kwargs: Additional arguments which are passed here.

        Returns:
            List of test results.
        """
        results = []
        input_metadata = kwargs.get("input_metadata", None)

        # telemetry call for tracking test runs
        track_e2e_test_run(input_test_cases, input_fixtures, input_metadata)

        for test_case in input_test_cases:
            test_case_name = test_case.name.replace(" ", "_")
            # Add the name of the file and the current test case name being
            # executed in order to properly retrieve stub custom action
            if self.agent.endpoints and self.agent.endpoints.action:
                self.agent.endpoints.action.kwargs[TEST_FILE_NAME] = Path(
                    test_case.file
                ).name
                self.agent.endpoints.action.kwargs[TEST_CASE_NAME] = test_case_name

            sender_id = self.generate_sender_id(test_case.name)
            _, test_result = await self._process_test_case(
                test_case, sender_id, input_fixtures, input_metadata
            )

            results.append(test_result)

            coverage = kwargs.get("coverage", False)
            if coverage:
                tracker = await self.agent.tracker_store.retrieve(sender_id)
                if tracker:
                    test_result.tested_paths, test_result.tested_commands = (
                        self._get_tested_flow_paths_and_commands(
                            tracker.events, test_result
                        )
                    )

            if fail_fast and not test_result.pass_status:
                break

        return results

    async def _run_test_case(
        self,
        sender_id: str,
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
        test_case: TestCase,
    ) -> TEST_TURNS_TYPE:
        collector = CollectingOutputChannel()

        if input_fixtures:
            test_fixtures = self.filter_fixtures_for_test_case(
                test_case, input_fixtures
            )
            await self.set_up_fixtures(test_fixtures, sender_id)

        test_case_metadata = None
        if input_metadata:
            test_case_metadata = self.filter_metadata_for_input(
                test_case.metadata_name, input_metadata
            )

        return await self.run_prediction_loop(
            collector,
            test_case.steps,
            sender_id,
            test_case_metadata,
            input_metadata,
        )

    @staticmethod
    def generate_sender_id(test_case_name: str) -> str:
        # add timestamp suffix to ensure sender_id is unique
        return f"{test_case_name}_{datetime.datetime.now()}"

    async def _process_test_case(
        self,
        test_case: TestCase,
        sender_id: str,
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
    ) -> Tuple[TEST_TURNS_TYPE, TestResult]:
        """Runs a single test case and returns the test turns and result."""
        test_turns = await self._run_test_case(
            sender_id, input_fixtures, input_metadata, test_case
        )

        if not test_case.uses_assertions():
            test_result = self.generate_test_result(test_turns, test_case)
        else:
            test_result = await self.run_assertions(
                sender_id, test_case, input_metadata
            )

        return test_turns, test_result

    async def run_tests_for_fine_tuning(
        self,
        input_test_cases: List[TestCase],
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
    ) -> List[Conversation]:
        """Runs the test cases for fine-tuning.

        Converts passing test cases into conversation objects containing the
        prompts and llm commands per user message.

        Args:
            input_test_cases: Input test cases.
            input_fixtures: Input fixtures.
            input_metadata: Input metadata.

        Returns:
            List of conversations.
        """
        import rasa.llm_fine_tuning.annotation_module

        conversations = []

        for i in tqdm(range(len(input_test_cases))):
            test_case = input_test_cases[i]
            sender_id = self.generate_sender_id(test_case.name)

            test_turns, test_result = await self._process_test_case(
                test_case, sender_id, input_fixtures, input_metadata
            )

            if not test_result.pass_status:
                structlogger.warning(
                    "annotation_module.skip_test_case.failing_e2e_test",
                    test_case=test_case.name,
                    file=test_case.file,
                )
                continue

            tracker = await self.agent.tracker_store.retrieve(sender_id)
            conversation = rasa.llm_fine_tuning.annotation_module.generate_conversation(
                test_turns, test_case, tracker, test_case.uses_assertions()
            )

            if conversation:
                conversations.append(conversation)

        return conversations

    async def run_tests_to_convert_tests_to_du_tests(
        self,
        input_test_cases: List[TestCase],
        input_fixtures: List[Fixture],
        input_metadata: Optional[List[Metadata]],
        converting_method: Callable[
            [TEST_TURNS_TYPE, TestCase, bool, bool],
            Optional[DialogueUnderstandingTestCase],
        ],
    ) -> Tuple[
        List[DialogueUnderstandingTestCase], List[DialogueUnderstandingTestCase]
    ]:
        """Runs the test cases to convert them into dialogue understanding tests.

        Converts test cases into dialogue understanding test cases.

        Args:
            input_test_cases: Input test cases.
            input_fixtures: Input fixtures.
            input_metadata: Input metadata.
            converting_method: The method to convert the e2e test case into a
              dialogue understanding test case.

        Returns:
            List of ready dialogue understanding test cases and list of
            dialogue understanding test cases to review.
        """
        ready_du_test_cases = []
        to_review_du_test_cases = []

        for i in tqdm(range(len(input_test_cases))):
            test_case = input_test_cases[i]
            sender_id = self.generate_sender_id(test_case.name)

            test_turns, test_result = await self._process_test_case(
                test_case, sender_id, input_fixtures, input_metadata
            )

            du_test_case = converting_method(
                test_turns,
                test_case,
                test_case.uses_assertions(),
                test_result.pass_status,
            )

            if du_test_case:
                if test_result.pass_status:
                    ready_du_test_cases.append(du_test_case)
                else:
                    to_review_du_test_cases.append(du_test_case)

        return ready_du_test_cases, to_review_du_test_cases

    @staticmethod
    def _action_server_is_reachable(
        endpoints: AvailableEndpoints, module: str = "e2e_test_runner"
    ) -> None:
        """Calls the action server health endpoint."""
        if not endpoints.action:
            structlogger.debug(
                f"{module}._action_server_is_reachable",
                message="No action endpoint configured. Skipping the health check "
                "of the action server.",
            )
            return

        if endpoints.action.actions_module:
            structlogger.debug(
                f"{module}._action_server_is_reachable",
                message="Rasa server is configured to run custom actions directly. "
                "Skipping the health check of the action server.",
            )
            return

        if not endpoints.action.url:
            structlogger.debug(
                f"{module}._action_server_is_reachable",
                message="Action endpoint URL is not defined in the endpoint "
                "configuration.",
            )
            return

        structlogger.debug(
            f"{module}._action_server_is_reachable",
            message="Detected action URL in the endpoint configuration.\n"
            f"Action Server URL: {endpoints.action.url}\n"
            "Sending a health request to the action endpoint.",
        )
        url = urlparse(endpoints.action.url)
        # replace /<path> with just /health
        url = url._replace(path="/health").geturl()  # type: ignore[assignment]
        try:
            response = requests.get(url, timeout=3)
        except requests.exceptions.ConnectionError as error:
            raise RasaException(
                "Action endpoint could not be reached. "
                "Actions server URL is defined in your endpoint configuration as "
                f"'{endpoints.action.url}'.\n"
                "Please make sure your action server is running and properly "
                "configured. Since running tests without a action server may "
                f"lead to unpredictable results.\n{error}"
            )

        if response.status_code != 200:
            raise RasaException(
                "Action endpoint is responding, but health status responded with "
                f"code {response.status_code}. Make sure your action server"
                " is properly configured and that the '/health' endpoint is available."
            )

        structlogger.debug(
            f"{module}._action_server_is_reachable",
            message="Action endpoint has responded successfully.\n"
            f"Response message: {response.text}\n"
            f"Response status code: {response.status_code}.",
        )

    def _get_tested_flow_paths_and_commands(
        self, events: List[Event], test_result: TestResult
    ) -> Tuple[Optional[List[FlowPath]], Dict[str, Dict[str, int]]]:
        """Extract tested paths and commands from dialog events.

        A flow path consists of bot utterances and custom actions.

        Args:
            events: The list of dialog events.
            test_result: The result of the test incl. the pass status.

        Returns:
            Tuple[flow_paths: Optional[List[FlowPath]], tested_commands:
            Dict[str, Dict[str, int]]], where tested_commands is a
            dictionary like
            {"flow1": {"set slot": 5, "clarify": 1}, "flow2": {"set slot": 3}}
        """
        tested_paths = []
        # we want to create a flow path per flow the e2e test covers
        # as an e2e test can cover multiple flows, we might end up creating
        # multiple flow paths
        _tested_commands: DefaultDict[str, DefaultDict[str, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        flow_paths_stack = []

        for event in events:
            if isinstance(event, FlowStarted) and not event.flow_id.startswith(
                RASA_DEFAULT_FLOW_PATTERN_PREFIX
            ):
                flow_paths_stack.append(FlowPath(event.flow_id))

            elif (
                isinstance(event, FlowCompleted)
                and len(flow_paths_stack) > 0
                and event.flow_id == flow_paths_stack[-1].flow
            ):
                # flow path is completed as the flow ended
                tested_paths.append(flow_paths_stack.pop())

            elif isinstance(event, BotUttered):
                if (
                    flow_paths_stack
                    and STEP_ID_METADATA_KEY in event.metadata
                    and ACTIVE_FLOW_METADATA_KEY in event.metadata
                ):
                    flow_paths_stack[-1].nodes.append(self._create_path_node(event))

            elif isinstance(event, ActionExecuted):
                # we are only interested in custom actions
                if (
                    flow_paths_stack
                    and self.agent.domain
                    and self.agent.domain.is_custom_action(event.action_name)
                    and STEP_ID_METADATA_KEY in event.metadata
                    and ACTIVE_FLOW_METADATA_KEY in event.metadata
                ):
                    flow_paths_stack[-1].nodes.append(self._create_path_node(event))

            # Time to gather tested commands
            elif isinstance(event, UserUttered):
                if event.parse_data and COMMANDS in event.parse_data:
                    commands = [
                        command["command"] for command in event.parse_data[COMMANDS]
                    ]
                    current_flow = (
                        flow_paths_stack[-1].flow if flow_paths_stack else "no_flow"
                    )
                    for command in commands:
                        _tested_commands[current_flow][command] += 1

        # It might be that an e2e test stops before a flow was completed.
        # Add the remaining flow paths to the tested paths list.
        while len(flow_paths_stack) > 0:
            tested_paths.append(flow_paths_stack.pop())

        # Convert _tested_commands to normal dicts
        tested_commands = {key: dict(value) for key, value in _tested_commands.items()}  # type: Dict[str, Dict[str, int]]

        return tested_paths, tested_commands

    @staticmethod
    def _create_path_node(event: Event) -> PathNode:
        flow_id = event.metadata[ACTIVE_FLOW_METADATA_KEY]
        step_id = event.metadata[STEP_ID_METADATA_KEY]
        return PathNode(step_id=step_id, flow=flow_id)



from dataclasses import dataclass, field
from typing import List


@dataclass
class RephrasedUserMessage:
    original_user_message: str
    rephrasings: List[str]
    failed_rephrasings: List[str] = field(default_factory=list)
    passed_rephrasings: List[str] = field(default_factory=list)



from __future__ import annotations

from dataclasses import dataclass
from typing import Text

from rasa.shared.core.flows.flow_step_links import FlowStepLinks, StaticFlowStepLink
from rasa.shared.core.flows.steps.constants import (
    CONTINUE_STEP_PREFIX,
    UNSET_FLOW_STEP_ID,
)
from rasa.shared.core.flows.steps.internal import InternalFlowStep


@dataclass
class ContinueFlowStep(InternalFlowStep):
    """A flow step that is dynamically introduced to jump to other flow steps."""

    def __init__(self, flow_id: Text, target_step_id: str) -> None:
        """Initializes a continue-step flow step."""
        super().__init__(
            idx=UNSET_FLOW_STEP_ID,
            custom_id=CONTINUE_STEP_PREFIX + target_step_id,
            description=None,
            metadata={},
            # The continue step links to the step that should be continued.
            # The flow policy in a sense only "runs" the logic of a step
            # when it transitions to that step, once it is there it will use
            # the next link to transition to the next step. This means that
            # if we want to "re-run" a step, we need to link to it again.
            # This is why the continue step links to the step that should be
            # continued.
            next=FlowStepLinks(links=[StaticFlowStepLink(target_step_id)]),
            flow_id=flow_id,
        )

    @staticmethod
    def continue_step_for_id(step_id: str) -> str:
        return CONTINUE_STEP_PREFIX + step_id



import asyncio
from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from werkzeug.local import LocalProxy


# Per voice session data
# This is similar to how flask makes the "request" object available as a global variable
# It's a "global" variable that is local to an async task (i.e. websocket session)
@dataclass
class CallState:
    is_user_speaking: bool = False
    is_bot_speaking: bool = False
    silence_timeout_watcher: Optional[asyncio.Task] = None
    silence_timeout: Optional[float] = None
    latest_bot_audio_id: Optional[str] = None
    should_hangup: bool = False
    connection_failed: bool = False

    # Generic field for channel-specific state data
    channel_data: Dict[str, Any] = field(default_factory=dict)


_call_state: ContextVar[CallState] = ContextVar("call_state")
call_state = LocalProxy(_call_state)



import logging
import os
from typing import Dict, List, Optional, Text, Union

import rasa.shared.core.flows.yaml_flows_io
import rasa.shared.data
import rasa.shared.utils.io
from rasa.shared.core.domain import Domain, InvalidDomain
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
    YAMLStoryReader,
)
from rasa.shared.core.training_data.structures import StoryGraph
from rasa.shared.importers import utils
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.utils.common import cached_method
from rasa.shared.utils.yaml import read_model_configuration

logger = logging.getLogger(__name__)


class RasaFileImporter(TrainingDataImporter):
    """Default `TrainingFileImporter` implementation."""

    def __init__(
        self,
        config_file: Optional[Text] = None,
        domain_path: Optional[Text] = None,
        training_data_paths: Optional[Union[List[Text], Text]] = None,
        expand_env_vars: bool = True,
    ):
        self.expand_env_vars = expand_env_vars
        self._domain_path = domain_path

        self._nlu_files = rasa.shared.data.get_data_files(
            training_data_paths, rasa.shared.data.is_nlu_file
        )
        self._story_files = rasa.shared.data.get_data_files(
            training_data_paths, YAMLStoryReader.is_stories_file
        )
        self._flow_files = rasa.shared.data.get_data_files(
            training_data_paths, rasa.shared.core.flows.yaml_flows_io.is_flows_file
        )
        self._conversation_test_files = rasa.shared.data.get_data_files(
            training_data_paths, YAMLStoryReader.is_test_stories_file
        )

        self.config_file = config_file

    @cached_method
    def get_config(self) -> Dict:
        """Retrieves model config (see parent class for full docstring)."""
        if not self.config_file or not os.path.exists(self.config_file):
            logger.debug("No configuration file was provided to the RasaFileImporter.")
            return {}

        config = read_model_configuration(
            self.config_file, expand_env_vars=self.expand_env_vars
        )
        return config

    def get_config_file_for_auto_config(self) -> Optional[Text]:
        """Returns config file path for auto-config only if there is a single one."""
        return self.config_file

    @cached_method
    def get_stories(self, exclusion_percentage: Optional[int] = None) -> StoryGraph:
        """Retrieves training stories / rules (see parent class for full docstring)."""
        return utils.story_graph_from_paths(
            self._story_files, self.get_domain(), exclusion_percentage
        )

    @cached_method
    def get_flows(self) -> FlowsList:
        """Retrieves training stories / rules (see parent class for full docstring)."""
        return utils.flows_from_paths(self._flow_files)

    @cached_method
    def get_conversation_tests(self) -> StoryGraph:
        """Retrieves conversation test stories (see parent class for full docstring)."""
        return utils.story_graph_from_paths(
            self._conversation_test_files, self.get_domain()
        )

    @cached_method
    def get_nlu_data(self, language: Optional[Text] = "en") -> TrainingData:
        """Retrieves NLU training data (see parent class for full docstring)."""
        return utils.training_data_from_paths(self._nlu_files, language)

    @cached_method
    def get_domain(self) -> Domain:
        """Retrieves model domain (see parent class for full docstring)."""
        domain = Domain.empty()

        # If domain path is None, return an empty domain
        if not self._domain_path:
            return domain
        try:
            domain = Domain.load(self._domain_path)
        except InvalidDomain as e:
            rasa.shared.utils.io.raise_warning(
                f"Loading domain from '{self._domain_path}' failed. Using "
                f"empty domain. Error: '{e}'"
            )

        return domain



from __future__ import annotations

from typing import Any, Dict, Optional, Text

from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.domain import Domain
from rasa.shared.exceptions import InvalidConfigException
from rasa.shared.importers.importer import TrainingDataImporter


class DomainProvider(GraphComponent):
    """Provides domain during training and inference time."""

    def __init__(
        self,
        model_storage: ModelStorage,
        resource: Resource,
        domain: Optional[Domain] = None,
    ) -> None:
        """Creates domain provider."""
        self._model_storage = model_storage
        self._resource = resource
        self._domain = domain

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> DomainProvider:
        """Creates component (see parent class for full docstring)."""
        return cls(model_storage, resource)

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> DomainProvider:
        """Creates provider using a persisted version of itself."""
        with model_storage.read_from(resource) as resource_directory:
            domain = Domain.from_path(resource_directory)
        return cls(model_storage, resource, domain)

    def _persist(self, domain: Domain) -> None:
        """Persists domain to model storage."""
        with self._model_storage.write_to(self._resource) as resource_directory:
            domain.persist(resource_directory / "domain.yml")

    def provide_train(self, importer: TrainingDataImporter) -> Domain:
        """Provides domain from training data during training."""
        domain = importer.get_domain()
        self._persist(domain)
        return domain

    def provide_inference(self) -> Domain:
        """Provides the domain during inference."""
        if self._domain is None:
            # This can't really happen but if it happens then we fail early
            raise InvalidConfigException(
                "No domain was found. This is required for "
                "making model predictions. Please make sure to "
                "provide a valid domain during training."
            )
        return self._domain



import argparse
import os
import sys
from collections import defaultdict
from enum import Enum
from typing import List, Text

from rasa import telemetry
from rasa.api import train
from rasa.cli import SubParsersAction
from rasa.cli.inspect import inspect
from rasa.shared.constants import (
    DEFAULT_CONFIG_PATH,
    DEFAULT_DATA_PATH,
    DEFAULT_DOMAIN_PATH,
    DEFAULT_MODELS_PATH,
    DOCS_BASE_URL,
)
from rasa.shared.utils.cli import print_error_and_exit, print_success


class ProjectTemplateName(Enum):
    """Enum of the different project templates."""

    DEFAULT = "default"
    TUTORIAL = "tutorial"
    CALM = "calm"

    def __str__(self) -> str:
        return self.value


template_domain_path = defaultdict(lambda: DEFAULT_DOMAIN_PATH)
template_domain_path[ProjectTemplateName.CALM] = "domain"


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all init parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    scaffold_parser = subparsers.add_parser(
        "init",
        parents=parents,
        help="Creates a new project, with example training data, "
        "actions, and config files.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    scaffold_parser.add_argument(
        "--no-prompt",
        action="store_true",
        help="Automatically choose default options for prompts and suppress warnings.",
    )
    scaffold_parser.add_argument(
        "--init-dir",
        default=None,
        help="Directory where your project should be initialized.",
    )
    scaffold_parser.add_argument(
        "--template",
        type=ProjectTemplateName,
        choices=list(ProjectTemplateName),
        default=ProjectTemplateName.DEFAULT,
        help="Select the template to use for the project.",
    )
    scaffold_parser.set_defaults(func=run)


def print_train_or_instructions(args: argparse.Namespace) -> None:
    """Train a model if the user wants to."""
    import questionary

    print_success("Finished creating project structure.")

    should_train = (
        questionary.confirm("Do you want to train an initial model? 💪🏽")
        .skip_if(args.no_prompt, default=True)
        .ask()
    )

    if should_train:
        print_success("Training an initial model...")
        training_result = train(
            template_domain_path[args.template],
            DEFAULT_CONFIG_PATH,
            DEFAULT_DATA_PATH,
            DEFAULT_MODELS_PATH,
        )
        args.model = training_result.model

        print_run_or_instructions(args)

    else:
        print_success(
            "No problem 👍🏼. You can also train a model later by going "
            "to the project directory and running 'rasa train'."
        )


def print_run_or_instructions(args: argparse.Namespace) -> None:
    import questionary

    from rasa.core import constants

    should_run = (
        questionary.confirm("Do you want to speak to the trained assistant? 🤖")
        .skip_if(args.no_prompt, default=False)
        .ask()
    )

    if should_run:
        # provide defaults for command line arguments
        attributes = [
            "endpoints",
            "credentials",
            "cors",
            "auth_token",
            "jwt_secret",
            "jwt_method",
            "enable_api",
            "remote_storage",
        ]
        for a in attributes:
            setattr(args, a, None)

        args.port = constants.DEFAULT_SERVER_PORT
        args.voice = False
        args.skip_yaml_validation = ["domain"]

        inspect(args)
    else:
        if args.no_prompt:
            print(
                "If you want to speak to the assistant, "
                "run 'rasa inspect' at any time inside "
                "the project directory."
            )
        else:
            print_success(
                "Ok 👍🏼. "
                "If you want to speak to the assistant, "
                "run 'rasa inspect' at any time inside "
                "the project directory."
            )


def init_project(args: argparse.Namespace, path: Text) -> None:
    """Inits project."""
    os.chdir(path)
    create_initial_project(".", args.template)
    print(f"Created project directory at '{os.getcwd()}'.")
    print_train_or_instructions(args)


def create_initial_project(
    path: Text, template: ProjectTemplateName = ProjectTemplateName.DEFAULT
) -> None:
    """Creates directory structure and templates for initial project."""
    from distutils.dir_util import copy_tree

    copy_tree(scaffold_path(template), path)


def scaffold_path(template: ProjectTemplateName) -> Text:
    import importlib_resources

    import rasa.cli.project_templates

    template_module = rasa.cli.project_templates.__name__
    return str(importlib_resources.files(template_module).joinpath(template.value))


def print_cancel() -> None:
    print_success("Ok. You can continue setting up by running 'rasa init' 🙋🏽‍♀️")
    sys.exit(0)


def _ask_create_path(path: Text) -> None:
    import questionary

    should_create = questionary.confirm(
        f"Path '{path}' does not exist 🧐. Create path?"
    ).ask()

    if should_create:
        try:
            os.makedirs(path)
        except (PermissionError, OSError, FileExistsError) as e:
            print_error_and_exit(
                f"Failed to create project path at '{path}'. " f"Error: {e}"
            )
    else:
        print_success(
            "Ok, will exit for now. You can continue setting up by "
            "running 'rasa init' again 🙋🏽‍♀️"
        )
        sys.exit(0)


def _ask_overwrite(path: Text) -> None:
    import questionary

    overwrite = questionary.confirm(
        "Directory '{}' is not empty. Continue?".format(os.path.abspath(path))
    ).ask()
    if not overwrite:
        print_cancel()


def run(args: argparse.Namespace) -> None:
    import questionary

    print_success("Welcome to Rasa! 🤖\n")
    print(
        "By installing and using this software, you agree to be bound by the terms "
        "and conditions of the Developer Terms available at "
        "https://rasa.com/developer-terms. "
        "Please review the Developer Terms carefully before proceeding.\n"
    )
    if args.no_prompt:
        print(
            f"To get started quickly, an "
            f"initial project will be created.\n"
            f"If you need some help, check out "
            f"the documentation at {DOCS_BASE_URL}.\n"
        )
    else:
        print(
            f"To get started quickly, an "
            f"initial project will be created.\n"
            f"If you need some help, check out "
            f"the documentation at {DOCS_BASE_URL}.\n"
            f"Now let's start! 👇🏽\n"
        )

    if args.init_dir is not None:
        path = args.init_dir
    else:
        path = (
            questionary.text(
                "Please enter a path where the project will be "
                "created [default: current directory]"
            )
            .skip_if(args.no_prompt, default="")
            .ask()
        )
        # set the default directory. we can't use the `default` property
        # in questionary as we want to avoid showing the "." in the prompt as the
        # initial value. users tend to overlook it and it leads to invalid
        # paths like: ".C:\mydir".
        # Can't use `if not path` either, as `None` will be handled differently (abort)
        if path == "":
            path = "."

    if args.no_prompt and not os.path.isdir(path):
        print_error_and_exit(f"Project init path '{path}' not found.")

    if path and not os.path.isdir(path):
        _ask_create_path(path)

    if path is None or not os.path.isdir(path):
        print_cancel()

    if not args.no_prompt and len(os.listdir(path)) > 0:
        _ask_overwrite(path)

    telemetry.track_project_init(path)

    init_project(args, path)



import logging
import os
from pathlib import Path
from socket import SOCK_DGRAM, SOCK_STREAM
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Text, Tuple, Union

import numpy as np
import structlog
from sanic import Sanic

import rasa.cli.utils as cli_utils
import rasa.shared.utils.io
from rasa.constants import DEFAULT_SANIC_WORKERS, ENV_SANIC_WORKERS
from rasa.core.constants import (
    ACTIVE_FLOW_METADATA_KEY,
    DOMAIN_GROUND_TRUTH_METADATA_KEY,
    STEP_ID_METADATA_KEY,
    UTTER_SOURCE_METADATA_KEY,
)
from rasa.core.lock_store import InMemoryLockStore, LockStore, RedisLockStore
from rasa.shared.constants import DEFAULT_ENDPOINTS_PATH, TCP_PROTOCOL
from rasa.shared.core.constants import SlotMappingType
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.utils.endpoints import (
    EndpointConfig,
    read_endpoint_config,
    read_property_config_from_endpoints_file,
)
from rasa.utils.io import write_yaml

if TYPE_CHECKING:
    from rasa.core.nlg import NaturalLanguageGenerator
    from rasa.shared.core.domain import Domain
    from rasa.shared.core.flows.flows_list import FlowsList

structlogger = structlog.get_logger()


def configure_file_logging(
    logger_obj: logging.Logger,
    log_file: Optional[Text],
    use_syslog: Optional[bool],
    syslog_address: Optional[Text] = None,
    syslog_port: Optional[int] = None,
    syslog_protocol: Optional[Text] = None,
) -> None:
    """Configure logging to a file.

    Args:
        logger_obj: Logger object to configure.
        log_file: Path of log file to write to.
        use_syslog: Add syslog as a logger.
        syslog_address: Adress of the syslog server.
        syslog_port: Port of the syslog server.
        syslog_protocol: Protocol with the syslog server
    """
    if use_syslog:
        formatter = logging.Formatter(
            "%(asctime)s [%(levelname)-5.5s] [%(process)d]" " %(message)s"
        )
        socktype = SOCK_STREAM if syslog_protocol == TCP_PROTOCOL else SOCK_DGRAM
        syslog_handler = logging.handlers.SysLogHandler(
            address=(syslog_address, syslog_port), socktype=socktype
        )
        syslog_handler.setLevel(logger_obj.level)
        syslog_handler.setFormatter(formatter)
        logger_obj.addHandler(syslog_handler)
    if log_file:
        formatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")
        file_handler = logging.FileHandler(
            log_file, encoding=rasa.shared.utils.io.DEFAULT_ENCODING
        )
        file_handler.setLevel(logger_obj.level)
        file_handler.setFormatter(formatter)
        logger_obj.addHandler(file_handler)


def one_hot(hot_idx: int, length: int, dtype: Optional[Text] = None) -> np.ndarray:
    """Create a one-hot array.

    Args:
        hot_idx: Index of the hot element.
        length: Length of the array.
        dtype: ``numpy.dtype`` of the array.

    Returns:
        One-hot array.
    """
    if hot_idx >= length:
        raise ValueError(
            "Can't create one hot. Index '{}' is out of range (length '{}')".format(
                hot_idx, length
            )
        )
    r = np.zeros(length, dtype)
    r[hot_idx] = 1
    return r


def dump_obj_as_yaml_to_file(
    filename: Union[Text, Path], obj: Any, should_preserve_key_order: bool = False
) -> None:
    """Writes `obj` to the filename in YAML repr.

    Args:
        filename: Target filename.
        obj: Object to dump.
        should_preserve_key_order: Whether to preserve key order in `obj`.
    """
    write_yaml(obj, filename, should_preserve_key_order=should_preserve_key_order)


def list_routes(app: Sanic) -> Dict[Text, Text]:
    """List all the routes of a sanic application. Mainly used for debugging."""
    from urllib.parse import unquote

    output = {}

    def find_route(suffix: Text, path: Text) -> Optional[Text]:
        for name, (uri, _) in app.router.routes_names.items():
            if name.split(".")[-1] == suffix and uri == path:
                return name
        return None

    for route in app.router.routes:
        endpoint = route.parts
        if endpoint[:-1] in app.router.routes_all and endpoint[-1] == "/":
            continue

        options = {}
        for arg in route._params:
            options[arg] = f"[{arg}]"

        name = route.name.replace("rasa_server.", "")
        methods = ",".join(route.methods)

        full_endpoint = "/" + "/".join(endpoint)
        line = unquote(f"{full_endpoint:50s} {methods:30s} {name}")
        output[name] = line

    url_table = "\n".join(output[url] for url in sorted(output))
    structlogger.debug(
        "server.routes", event_info=f"Available web server routes: \n{url_table}"
    )

    return output


def extract_args(
    kwargs: Dict[Text, Any], keys_to_extract: Set[Text]
) -> Tuple[Dict[Text, Any], Dict[Text, Any]]:
    """Go through the kwargs and filter out the specified keys.

    Return both, the filtered kwargs as well as the remaining kwargs.
    """
    remaining = {}
    extracted = {}
    for k, v in kwargs.items():
        if k in keys_to_extract:
            extracted[k] = v
        else:
            remaining[k] = v

    return extracted, remaining


def is_limit_reached(num_messages: int, limit: Optional[int]) -> bool:
    """Determine whether the number of messages has reached a limit.

    Args:
        num_messages: The number of messages to check.
        limit: Limit on the number of messages.

    Returns:
        `True` if the limit has been reached, otherwise `False`.
    """
    return limit is not None and num_messages >= limit


class AvailableEndpoints:
    """Collection of configured endpoints."""

    _instance = None

    @classmethod
    def read_endpoints(cls, endpoint_file: Text) -> "AvailableEndpoints":
        """Read the different endpoints from a yaml file."""
        nlg = read_endpoint_config(endpoint_file, endpoint_type="nlg")
        nlu = read_endpoint_config(endpoint_file, endpoint_type="nlu")
        action = read_endpoint_config(endpoint_file, endpoint_type="action_endpoint")
        model = read_endpoint_config(endpoint_file, endpoint_type="models")
        tracker_store = read_endpoint_config(
            endpoint_file, endpoint_type="tracker_store"
        )
        lock_store = read_endpoint_config(endpoint_file, endpoint_type="lock_store")
        event_broker = read_endpoint_config(endpoint_file, endpoint_type="event_broker")
        vector_store = read_endpoint_config(endpoint_file, endpoint_type="vector_store")
        model_groups = read_property_config_from_endpoints_file(
            endpoint_file, property_name="model_groups"
        )

        return cls(
            nlg,
            nlu,
            action,
            model,
            tracker_store,
            lock_store,
            event_broker,
            vector_store,
            model_groups,
        )

    def __init__(
        self,
        nlg: Optional[EndpointConfig] = None,
        nlu: Optional[EndpointConfig] = None,
        action: Optional[EndpointConfig] = None,
        model: Optional[EndpointConfig] = None,
        tracker_store: Optional[EndpointConfig] = None,
        lock_store: Optional[EndpointConfig] = None,
        event_broker: Optional[EndpointConfig] = None,
        vector_store: Optional[EndpointConfig] = None,
        model_groups: Optional[List[Dict[str, Any]]] = None,
    ) -> None:
        """Create an `AvailableEndpoints` object."""
        self.model = model
        self.action = action
        self.nlu = nlu
        self.nlg = nlg
        self.tracker_store = tracker_store
        self.lock_store = lock_store
        self.event_broker = event_broker
        self.vector_store = vector_store
        self.model_groups = model_groups

    @classmethod
    def get_instance(
        cls, endpoint_file: Optional[Text] = DEFAULT_ENDPOINTS_PATH
    ) -> "AvailableEndpoints":
        """Get the singleton instance of AvailableEndpoints."""
        # Ensure that the instance is initialized only once.
        if cls._instance is None:
            cls._instance = cls.read_endpoints(endpoint_file)
        return cls._instance

    @classmethod
    def reset_instance(cls) -> None:
        cls._instance = None


def read_endpoints_from_path(
    endpoints_path: Optional[Union[Path, Text]] = None,
) -> AvailableEndpoints:
    """Get `AvailableEndpoints` object from specified path.

    Args:
        endpoints_path: Path of the endpoints file to be read. If `None` the
            default path for that file is used (`endpoints.yml`).

    Returns:
        `AvailableEndpoints` object read from endpoints file.

    """
    endpoints_config_path = cli_utils.get_validated_path(
        endpoints_path, "endpoints", DEFAULT_ENDPOINTS_PATH, True
    )
    return AvailableEndpoints.get_instance(endpoints_config_path)


def _lock_store_is_multi_worker_compatible(
    lock_store: Union[EndpointConfig, LockStore, None],
) -> bool:
    if isinstance(lock_store, InMemoryLockStore):
        return False

    if isinstance(lock_store, RedisLockStore):
        return True

    # `lock_store` is `None` or `EndpointConfig`
    return (
        lock_store is not None
        and isinstance(lock_store, EndpointConfig)
        and lock_store.type != "in_memory"
    )


def number_of_sanic_workers(lock_store: Union[EndpointConfig, LockStore, None]) -> int:
    """Get the number of Sanic workers to use in `app.run()`.

    If the environment variable constants.ENV_SANIC_WORKERS is set and is not equal to
    1, that value will only be permitted if the used lock store is not the
    `InMemoryLockStore`.
    """

    def _log_and_get_default_number_of_workers() -> int:
        structlogger.debug(
            "server.worker.set_count",
            number_of_workers=DEFAULT_SANIC_WORKERS,
            event_info=f"Using the default number of Sanic workers "
            f"({DEFAULT_SANIC_WORKERS}).",
        )
        return DEFAULT_SANIC_WORKERS

    try:
        env_value = int(os.environ.get(ENV_SANIC_WORKERS, DEFAULT_SANIC_WORKERS))
    except ValueError:
        structlogger.error(
            "server.worker.set_count.error",
            number_of_workers=os.environ[ENV_SANIC_WORKERS],
            event_info=f"Cannot convert environment variable `{ENV_SANIC_WORKERS}` "
            f"to int ('{os.environ[ENV_SANIC_WORKERS]}').",
        )
        return _log_and_get_default_number_of_workers()

    if env_value == DEFAULT_SANIC_WORKERS:
        return _log_and_get_default_number_of_workers()

    if env_value < 1:
        structlogger.warning(
            "server.worker.set_count.error_less_than_one",
            number_of_workers=env_value,
            event_info=f"Cannot set number of Sanic workers to the desired value "
            f"({env_value}). The number of workers must be at least 1.",
        )
        return _log_and_get_default_number_of_workers()

    if _lock_store_is_multi_worker_compatible(lock_store):
        structlogger.debug(
            "server.worker.set_count.success",
            event_info=f"Using {env_value} Sanic workers.",
            num_workers=env_value,
        )
        return env_value

    structlogger.warning(
        "server.worker.set_count.error_no_lock_store",
        event_info=f"Unable to assign desired number of Sanic workers ({env_value}) as "
        f"no `RedisLockStore` or custom `LockStore` endpoint "
        f"configuration has been found.",
        num_workers=env_value,
    )
    return _log_and_get_default_number_of_workers()


def add_bot_utterance_metadata(
    message: Dict[str, Any],
    domain_response_name: str,
    nlg: "NaturalLanguageGenerator",
    domain: "Domain",
    tracker: Optional[DialogueStateTracker],
) -> Dict[str, Any]:
    """Add metadata to the bot message."""
    message["utter_action"] = domain_response_name

    utter_source = message.get(UTTER_SOURCE_METADATA_KEY)
    if utter_source is None:
        utter_source = nlg.__class__.__name__
        message[UTTER_SOURCE_METADATA_KEY] = utter_source

    if tracker:
        message[ACTIVE_FLOW_METADATA_KEY] = tracker.active_flow
        message[STEP_ID_METADATA_KEY] = tracker.current_step_id

    if utter_source in ["IntentlessPolicy", "ContextualResponseRephraser"]:
        message[DOMAIN_GROUND_TRUTH_METADATA_KEY] = [
            response.get("text")
            for response in domain.responses.get(domain_response_name, [])
            if response.get("text") is not None
        ]

    return message


def should_force_slot_filling(
    tracker: Optional[DialogueStateTracker], flows: "FlowsList"
) -> Tuple[bool, Optional[str]]:
    """Check if the flow should force slot filling.

    This is only valid when the flow is at a collect information step which
    has set `force_slot_filling` to true and the slot has a valid `from_text` mapping.

    Args:
        tracker: The dialogue state tracker.
        flows: The list of flows.

    Returns:
        A tuple of a boolean indicating if the flow should force slot filling
        and the name of the slot if applicable.
    """
    from rasa.dialogue_understanding.processor.command_processor import (
        find_updated_flows,
        get_current_collect_step,
    )

    if tracker is None:
        structlogger.debug(
            "slot.force_slot_filling.no_found_tracker",
            event_info="Tracker is None. Cannot force slot filling.",
        )
        return False, None

    updated_flows = find_updated_flows(tracker, flows)
    if updated_flows:
        structlogger.debug(
            "slot.force_slot_filling.running_flows_were_updated",
            updated_flow_ids=updated_flows,
        )
        return False, None

    stack = tracker.stack
    step = get_current_collect_step(stack, flows)
    if step is None or not step.force_slot_filling:
        return False, None

    slot_name = step.collect
    slot = tracker.slots.get(slot_name)

    if not slot:
        structlogger.debug(
            "slot.force_slot_filling.error",
            event_info=f"Slot '{slot_name}' not found in tracker. "
            f"Cannot force slot filling. "
            f"Please check if the slot is defined in the domain.",
        )
        return False, None

    for slot_mapping in slot.mappings:
        if slot_mapping.type == SlotMappingType.FROM_TEXT:
            return True, slot_name

    return False, None
